"""
The entire BEEP CLI.

Guiding principles of how the CLI works:
 - Errors are thrown (stderr) if the *command itself* is bad.
 - Errors on individual files are caught during processing and
   reported via logging or status json.

The main "beep" command with no subcommand can specify where
and how to log all output and results (i.e., reporting).

The subcommands themselves specify where and how to process
actual BEEP operations such as structuring, featurization,
protocol generation, and running models.

"""

import os
import ast
import sys
import time
import copy
import pprint
import fnmatch
import hashlib
import logging
import datetime
import traceback
import importlib

import click
import numpy as np
import pandas as pd
from monty.serialization import dumpfn, loadfn

from beep import (
    logger,
    S3_CACHE,
    formatter_jsonl,
    __version__
)
from beep.structure.base import BEEPDatapath
from beep.structure.cli import auto_load, auto_load_processed
from beep.structure.validate import BEEPValidationError
from beep.features.base import (
    BEEPFeaturizer,
    BEEPFeaturizationError,
    BEEPFeatureMatrix,
)
from beep.features.core import (
    HPPCResistanceVoltageFeatures,
    DeltaQFastCharge,
    TrajectoryFastCharge,
    CycleSummaryStats,
    DiagnosticProperties,
    DiagnosticSummaryStats
)
from beep.features.intracell_losses import (
    IntracellCycles,
    IntracellFeatures
)
from beep.model import BEEPLinearModelExperiment
from beep.utils.s3 import list_s3_objects, download_s3_object
from beep.protocol.generate_protocol import generate_protocol_files_from_csv
from beep.protocol.generate_protocol import ProtocolException


CLICK_FILE = click.Path(file_okay=True, dir_okay=False, writable=False, readable=True)
CLICK_DIR = click.Path(file_okay=False, dir_okay=True, writable=True, readable=True)
STRUCTURED_SUFFIX = "_structured"
FEATURIZED_SUFFIX = "_featurized"


class ContextPersister:
    """
    Class to hold persisting objects for downstream BEEP tasks in the
    command line interface.

    Args:
        cwd (str): The current working directory.
        run_id (int): The current run id, if available, to associate with operation.
        tags ([str]): A list of tags to associate with an operation.
        output_status_json (str): Filename of status json to write.
        halt_on_error (bool): Whether to halt on error during an operation.
    """
    def __init__(
            self,
            cwd=None,
            run_id=None,
            tags=None,
            output_status_json=None,
            halt_on_error=None

    ):
        self.cwd = cwd
        self.run_id = run_id
        self.tags = tags
        self.output_status_json = output_status_json
        self.halt_on_error = halt_on_error


def add_suffix(full_path, output_dir, suffix, modified_ext=None):
    """Add structured filename suffixes and format new file paths.

    Args:
        full_path (str): Full path of the existing file.
        output_dir (str): Desired output directory of file.
        suffix (str): Suffix to append to filename before the extension.
        modified_ext (str): Extension to replace current file extension
            with. For example, ".json".

    Returns:
        (str): Renamed and reformatted file path.

    """
    basename = os.path.basename(full_path)
    stripped_basename, ext = os.path.splitext(basename)
    if modified_ext:
        ext = modified_ext
    new_basename = stripped_basename + suffix + ext
    return os.path.join(
        output_dir,
        new_basename
    )


def add_metadata_to_status_json(status_dict, run_id, tags):
    """Add some basic metadata to the status json.

    Args:
        status_dict (dict): Dictionary which will be written to status hson.
        run_id (int): Run id of this operation.
        tags ([str]): List of short string tags tagging an operation.

    Returns:
        (dict): Dictionary including BEEP metadata
    """
    metadata = {
        "beep_verison": __version__,
        "op_datetime_utc": datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
        "run_id": run_id,
        "tags": tags
    }
    status_dict["metadata"] = metadata
    return status_dict


def md5sum(filename):
    """
    Get md5 sum hash of a file.

    Args:
        filename (str): Name of the file.

    Returns:
        (str) Hash digest h.
    """
    with open(filename, "rb") as f:
        d = f.read()
        h = hashlib.md5(d).hexdigest()
    return h


@click.group(
    invoke_without_command=False,
    help="Base BEEP command."
)
@click.option(
    "--log-file",
    "-l",
    type=click.Path(file_okay=True, dir_okay=False, writable=True, readable=True),
    multiple=False,
    help="File to log formatted json to. Log will still be output in human "
         "readable form to stdout, but if --log-file is specified, it will "
         "be additionally logged to a jsonl (json-lines) formatted file.",
)
@click.option(
    "--run-id",
    "-r",
    type=click.INT,
    multiple=False,
    help="An integer run_id which can be optionally assigned to this run. "
         "It will be output in the metadata status json for any subcommand "
         "if the status json is enabled."
)
@click.option(
    "--tags",
    "-t",
    type=click.STRING,
    multiple=True,
    help="Add optional tags to the status json metadata. Can be later used for"
         "large-scale queries on database data about sets of BEEP runs. Example:"
         "'experiments_for_kristin'."
)
@click.option(
    '--output-status-json',
    '-s',
    type=CLICK_FILE,
    multiple=False,
    help="File to output with JSON info about the states of "
         "files which have had any beep subcommand operation"
         "run on them (e.g., structuring). Contains comprehensive"
         "info about the success of the operation for all files."
         "1 status json = 1 operation."
)
@click.option(
    '--halt-on-error',
    is_flag=True,
    default=False,
    help="Set to halt BEEP if critical featurization "
         "errors are encountered on any file with any featurizer. "
         "Otherwise, logs critical errors to the status json.",
)
@click.pass_context
def cli(ctx, log_file, run_id, tags, output_status_json, halt_on_error):
    """
    Base command for all BEEP subcommands. Sets CWD and persistent
    context.


    Args:
        log_file (str): The name of the log file to append log
            output to.
        run_id (int): Run identifier associated with this operation.
        tags ([str]): List of string tags to associate with this
            operation.
        output_status_json (str): File to write the status json to.
        halt_on_error (bool): If True, will halt the entire CLI
            operation if any exception occurs, even in suboperations.

    Returns:
        None
    """
    ctx.ensure_object(ContextPersister)
    cwd = os.path.abspath(os.getcwd())
    ctx.obj.cwd = cwd
    ctx.obj.tags = tags
    ctx.obj.run_id = run_id
    ctx.obj.output_status_json = output_status_json
    ctx.obj.halt_on_error = halt_on_error

    if log_file:
        hdlr = logging.FileHandler(log_file, "a")
        hdlr.setFormatter(formatter_jsonl)
        logger.addHandler(hdlr)


@cli.command(
    help="Structure and/or validate one or more files. Argument "
         "is a space-separated list of files or globs.",
    short_help="interpolate, validate, and standardize battery files"
)
@click.argument(
    'files',
    nargs=-1,
    type=CLICK_FILE,
)
@click.option(
    '--output-filenames',
    '-o',
    type=click.Path(),
    help="Filenames to write each input filename to. "
         "If not specified, auto-names each file by appending"
         "`-structured` before the file extension inside "
         "the current working dir.",
    multiple=True
)
@click.option(
    '--output-dir',
    '-d',
    type=CLICK_DIR,
    help="Directory to dump auto-named files to. Only works if"
         "--output-filenames is not specified."
)
@click.option(
    '--protocol-parameters-dir',
    '-p',
    type=CLICK_DIR,
    help="Directory of a protocol parameters files to use for "
         "auto-structuring. If not specified, BEEP cannot auto-"
         "structure. Use with --automatic."
)
@click.option(
    '--v-range',
    '-v',
    type=(click.FLOAT, click.FLOAT),
    help="Lower, upper bounds for voltage range for structuring. "
         "Overridden by auto-structuring if --automatic."
)
@click.option(
    '--resolution',
    '-r',
    type=click.INT,
    default=1000,
    help="Resolution for interpolation for structuring. Overridden "
         "by auto-structuring if --automatic."
)
@click.option(
    '--nominal-capacity',
    '-n',
    type=click.FLOAT,
    default=1.1,
    help="Nominal capacity to use for structuring. Overridden by "
         "auto-structuring if --automatic."
)
@click.option(
    '--full-fast-charge',
    '-f',
    type=click.FLOAT,
    default=0.8,
    help="Full fast charge threshold to use for structuring. "
         "Overridden by auto-structuring if --automatic."
)
@click.option(
    '--charge-axis',
    '-c',
    type=click.STRING,
    default='charge_capacity',
    help="Axis to use for charge step interpolation. Must be found "
         "inside the loaded dataframe. Can be used with --automatic."
)
@click.option(
    '--discharge-axis',
    '-x',
    type=click.STRING,
    default='voltage',
    help="Axis to use for discharge step interpolation. Must be "
         "found inside the loaded dataframe. Can be used with"
         "--automatic."
)
@click.option(
    '--s3-bucket',
    '-b',
    default=None,
    type=click.STRING,
    help="Expands file paths to include those in the s3 bucket specified. "
         "File paths specify s3 keys. Keys can be globbed/wildcarded. Paths "
         "matching local files will be prioritized over files with identical "
         "paths/globs in s3. Files will be downloaded to CWD."
)
@click.option(
    '--automatic',
    is_flag=True,
    default=False,
    help="If --protocol-parameters-path is specified, will "
         "automatically determine structuring parameters. Will override "
         "all manually set structuring parameters."
)
@click.option(
    '--validation-only',
    is_flag=True,
    default=False,
    help='Skips structuring, only validates files.'
)
@click.option(
    '--no-raw',
    is_flag=True,
    default=False,
    help="Does not save raw cycler data to disk. Saves disk space, but "
         "prevents files from being partially restructured."
)
@click.option(
    '--s3-use-cache',
    is_flag=True,
    default=False,
    help="Use s3 cache defined with environment variable BEEP_S3_CACHE "
         "instead of downloading files directly to the CWD."
)
@click.pass_context
def structure(
        ctx,
        files,
        output_filenames,
        output_dir,
        protocol_parameters_dir,
        v_range,
        resolution,
        nominal_capacity,
        full_fast_charge,
        charge_axis,
        discharge_axis,
        s3_bucket,
        automatic,
        validation_only,
        no_raw,
        s3_use_cache
):
    """
    Structure files using the beep cli.


    raw files ---> structured json files.



    Args:
        ctx (ContextPersister): A context-persisting object containing
            metadata and logging specifications for subcommands.
        files ([str]): Paths or globs for raw input files to structure.
        output_filenames ([str]): Corresponding output filenames for each
            input file to write structured output to.
        output_dir (str): Output directory to write auto-named structured files
            into. Overrides output_filenames.
        protocol_parameters_dir (str): Directory containing protocol parameters
            for automatically determining structuring parameters.
        v_range ((float, float)): Tuple of voltage floats to determine voltage range
            for interpolation during structuring. Overridden by automatic, if
            automatic specified.
        resolution (int): Number of points to interpolate during structuring. Overridden
            by automatic, if automatic specified.
        nominal_capacity (float): Nominal capacity of the cell to use during structuring.
            Overridden by automatic, if automatic specified.
        full_fast_charge (float): Fast charge threshold to use during structuring.
            Overridden by automatic, if automatic specified.
        charge_axis (str): Standardized name of the charge axis to interpolate with
            respect to. Overridden by automatic, if automatic specified.
        discharge_axis (str): Standardized name of the discharge axis to interpolate
            with respect to. Overridden by automatic, if automatic specified.
        s3_bucket (str): S3 bucket to check for files. Files matching keys in this
            bucket will automatically be downloaded to the CWD or S3 cache and then
            used for structuring.
        automatic (bool): If true, overrides other structuring parameters and uses
            the specified protocol_parameters_dir to automatically determine
            structuring parameters.
        validation_only (bool): If True, only validates files, does not structure them
            or produce structured outputs.
        no_raw (bool): If True, keeps only the structured data, does not retain
            the raw data in the structured output datapaths. WARNING: files structured
            with no_raw=True cannot be restructured without the raw file and operations
            are not reversible without access to the raw file.
        s3_use_cache (bool): If true, uses the beep S3 cache to store raw files rather
            than dumping them in the CWD.

    Returns:
        None

    """

    # download from s3 first, if needed
    if s3_bucket:
        logger.info(f"Fetching file list from s3 bucket {s3_bucket}...")
        s3_objs = list_s3_objects(s3_bucket)
        logger.info(f"Including {len(s3_objs)} available s3 objects in file match.")
        s3_keys = [o.key for o in s3_objs]

        # local files matching globs are pre-expanded by Click
        s3_keys_matched = []
        local_files = []
        for maybe_glob in files:
            # add direct matches
            if "*" not in maybe_glob:
                if maybe_glob in s3_keys:
                    s3_keys_matched.append(maybe_glob)
                else:
                    local_files.append(maybe_glob)
            else:
                # its a glob, and real local globs will
                # be pre-expanded by click, so the only
                # valid globs will be on s3. All remaining
                # globs are invalid/bad paths
                matching_files = fnmatch.filter(s3_keys, maybe_glob)
                if matching_files:
                    s3_keys_matched += matching_files
                else:
                    local_files.append(maybe_glob)

        logger.info(f"Found {len(s3_keys_matched)} matching files on s3")
        local_files_from_s3 = []
        for s3k in s3_keys_matched:
            s3k_basename = os.path.basename(s3k)
            pardir = S3_CACHE if s3_use_cache else ctx.obj.cwd
            pardir = os.path.abspath(pardir)
            if not os.path.exists(pardir):
                os.makedirs(pardir)
            s3k_local_fullname = os.path.join(pardir, s3k_basename)
            logger.info(f"Fetching {s3k} from {s3_bucket}")
            download_s3_object(s3_bucket, s3k, s3k_local_fullname)
            logger.info(f"Fetched s3 file {s3k_basename} to {s3k_local_fullname}")
            local_files_from_s3.append(s3k_local_fullname)
        files = local_files + local_files_from_s3

    files = [os.path.abspath(f) for f in files]

    for file in files:
        if not os.path.exists(file):
            raise FileNotFoundError(f"File '{file}' not found on filesystem!")
    n_files = len(files)

    logger.info(f"Structuring {n_files} files")

    # Output dir overrules output filenames
    if output_dir:
        # Use auto-naming in the output dir
        output_dir = os.path.abspath(output_dir)
        output_files = [
            add_suffix(f, output_dir, STRUCTURED_SUFFIX, modified_ext=".json.gz")
            for f in files
        ]

        if output_filenames:
            logger.warning(
                "Both --output-filenames and --output-dir were specified; "
                "defaulting to --output-dir with auto-naming."
            )
    else:
        if output_filenames:
            output_files = [os.path.abspath(f) for f in output_filenames]
            n_outputs = len(output_files)
            if n_files != n_outputs:
                raise ValueError(
                    f"Number of input files ({n_files}) does not match number "
                    f"of output filenames ({n_outputs})!"
                )
        else:
            # Use auto-naming in the cwd
            output_files = [
                add_suffix(f, ctx.obj.cwd, STRUCTURED_SUFFIX, modified_ext=".json.gz")
                for f in files
            ]

    if automatic and not protocol_parameters_dir:
        logger.warning(
            "--automatic was passed but no protocol parameters "
            "directory was specified! Default will be used."
            "Pass --protocol-parameters-dir to use autostructuring."
        )

    params = {
        "v_range": v_range,
        "resolution": resolution,
        "nominal_capacity": nominal_capacity,
        "full_fast_charge": full_fast_charge,
        "charge_axis": charge_axis,
        "discharge_axis": discharge_axis
    }

    status_json = {
        "op_type": "structure",
        "files": {}
    }

    log_prefix = "No file"
    for i, f in enumerate(files):
        op_result = {
            "validated": False,
            "validation_schema": None,
            "structured": False,
            "output": None,
            "traceback": None,
            "walltime": None,
            "raw_md5_chksum": None,
            "structuring_parameters": None,
            "processed_md5_chksum": None
        }

        t0 = time.time()
        try:
            log_prefix = f"File {i + 1} of {n_files}"
            logger.debug(f"Hashing file '{f}' to MD5")
            op_result["raw_md5_chksum"] = md5sum(f)

            logger.info(f"{log_prefix}: Reading raw file {f} from disk...")
            dp = auto_load(f)
            logger.info(f"{log_prefix}: Validating: {f} according to schema file '{dp.schema}'")
            op_result["validation_schema"] = dp.schema

            is_valid, validation_reason = dp.validate()
            op_result["validated"] = is_valid

            if not is_valid:
                raise BEEPValidationError(validation_reason)

            logger.info(f"File {i + 1} of {n_files}: Validated: {f}")

            if not validation_only:
                logger.info(f"{log_prefix}: Structuring: Read from {f}")
                if automatic:
                    dp.autostructure(
                        charge_axis=charge_axis,
                        discharge_axis=discharge_axis,
                        parameters_path=protocol_parameters_dir
                    )
                else:
                    dp.structure(**params)

                output_fname = output_files[i]
                dp.to_json_file(output_fname, omit_raw=no_raw)
                op_result["structured"] = True
                op_result["structuring_parameters"] = dp.structuring_parameters
                op_result["output"] = output_fname
                op_result["processed_md5_chksum"] = md5sum(output_fname)
                logger.info(f"{log_prefix}: Structured: Written to {output_fname}")

        except KeyboardInterrupt:
            logging.critical("Keyboard interrupt caught - exiting...")
            click.Context.exit(1)

        except BaseException:
            tbinfo = sys.exc_info()
            tbfmt = traceback.format_exception(*tbinfo)
            logger.error(f"{log_prefix}: Failed/invalid: ({tbinfo[0].__name__}): {f}")
            op_result["traceback"] = tbfmt

            if ctx.obj.halt_on_error:
                raise

        t1 = time.time()
        op_result["walltime"] = t1 - t0
        status_json["files"][f] = op_result

    # Generate the status report
    succeeded, failed, invalid = [], [], []

    for input_fname, op_result in status_json["files"].items():
        if op_result["validated"] and op_result["structured"]:
            succeeded.append(input_fname)
        elif op_result["validated"] and not op_result["structured"]:
            failed.append(input_fname)
        else:
            invalid.append(input_fname)

    logger.info(f"{'Validation' if validation_only else 'Structuring'} report:")

    logger.info(f"\t{'Structured' if validation_only else 'Succeeded'}: {len(succeeded)}/{n_files}")
    logger.info(f"\tInvalid: {len(invalid)}/{n_files}")
    for inv in invalid:
        logger.info(f"\t\t- {inv}")

    logger.info(f"\t{'Validated, not structured' if validation_only else 'Failed'}: {len(failed)}/{n_files}")
    for fail in failed:
        logger.info(f"\t\t- {fail}")

    status_json = add_metadata_to_status_json(status_json, ctx.obj.run_id, ctx.obj.tags)

    osj = ctx.obj.output_status_json
    if osj:
        dumpfn(status_json, osj)
        logger.info(f"Wrote status json file to {osj}")


@cli.command(
    help="Featurize one or more files. Argument "
         "is a space-separated list of files or globs. The same "
         "features are applied to each file. Naming of output "
         "files is done automatically, but the output directory "
         "can be specified.",
    short_help="create feature matrices from structured files"
)
@click.argument(
    'files',
    nargs=-1,
    type=CLICK_FILE,
)
@click.option(
    '--output-filename',
    '-o',
    type=CLICK_FILE,
    help="Filename to save entre feature matrix to. If not specified, output file"
         "will be named with FeatureMatrix-[timestamp].json.gz. If specified, "
         "overrides the output dir for saving the feature matrix to file."
)
@click.option(
    '--output-dir',
    '-d',
    type=CLICK_DIR,
    help="Directory to dump auto-named files to."
)
@click.option(
    '--featurize-with',
    "-f",
    default=["all_features"],
    multiple=True,
    type=click.STRING,
    help="Specify a featurizer to apply by class name, e.g. "
         "HPPCResistanceVoltageFeatures. To apply more than one "
         "featurizer, use multiple -f <FEATURIZER> commands. To apply sets of"
         "core BEEP featurizers, pass either 'all_features' for all features or"
         "'all_targets' for all targets (features which can be used as targets). "
         "Note if 'all_features' "
         "or 'all_targets' is passed, other -f featurizers will be ignored. All "
         "feautrizers are attempted to apply with default hyperparameters; "
         "to specify your own hyperparameters, use --featurize-with-hyperparams."
         "Classes from installed modules not in core BEEP can be "
         "specified with the class name in absolute import format, "
         "e.g., my_package.my_module.MyClass."
)
@click.option(
    "--featurize-with-hyperparams",
    "-h",
    multiple=True,
    help="Specify a featurizer to apply by class name with your own hyperparameters."
         "(such as parameter directories or specific values for hyperparameters"
         "for this featurizer), pass a dictionary in the format: "
         "'{\"FEATURIZER_NAME\": {\"HYPERPARAM1\": \"VALUE1\"...}}' including the "
         "single quotes around the outside and double quotes for internal strings."
         "Custom hyperparameters will be merged with default hyperparameters if the "
         "hyperparameter dictionary is underspecified.",
)
@click.option(
    "--save-intermediates",
    is_flag=True,
    help="Save the intermediate BEEPFeaturizers as json files. Filenames "
         "are autogenerated and saved in output-dir if specified; otherwise, "
         "intermediates are written to current working directory."
)
@click.pass_context
def featurize(
        ctx,
        files,
        output_filename,
        output_dir,
        featurize_with,
        featurize_with_hyperparams,
        save_intermediates,
):
    """
    Create feature matrices based on structured inputs for use with
    downstream models.

    Can also be used for creating various targets, such as expected
    cycles for given capacities, etc.


    structured json files ---> feature matrix json file

    AND

    structured json files ---> featurizer json files (optional)

    Args:
        ctx (ContextPersister): A context-persisting object containing
            metadata and logging specifications for subcommands.
        files ([str]): Filenames or globs of structured json/json.gz files to
            use for generating a new feature matrix.
        output_filename (str): Filename of the output feature matrix file.
        output_dir (str): Directory to dump auto-named file into. Will also
            dump intermediate files here if save_intermediates.
        featurize_with ([str]): Featurizer class names to use with default
            featurizer hyperparameters. Can include builtin "core" featurizers
            such as CycleSummaryStats or custom featurizers inheriting from
            BEEPFeaturizer by passing the fully qualified importable name
            of the class, e.g. "my_pkg.my_module.my_class".
        featurize_with_hyperparams (dict): More advanced hyperparameter specification
            for featurizers. Pass in one dictionary per featurizer with the
            featurizer class name as the sole key and a dictionary of parameters
            as the sole value. E.g., {"SomeFeaturizerClass": {"param1": 12}}.
            Underspecified hyperparameter dictionaries are merged with defaults.
        save_intermediates (bool): If True, will save the intermediate json files
            of BEEPFeaturizer objects used to construct the feature martix.

    Returns:
        None

    """
    files = [os.path.abspath(f) for f in files]

    n_files = len(files)
    output_dir = os.path.abspath(output_dir) if output_dir else ctx.obj.cwd

    logger.info(f"Featurizing {n_files} files")

    core_fclasses_feats = [
        HPPCResistanceVoltageFeatures,
        DeltaQFastCharge,
        DiagnosticSummaryStats,
        CycleSummaryStats,
    ]

    core_fclasses_targets = [
        TrajectoryFastCharge,
        DiagnosticProperties,
    ]
    native_fclasses = core_fclasses_feats + core_fclasses_targets + [IntracellCycles, IntracellFeatures]

    core_fclasses_feats_map = {fclass.__name__: fclass for fclass in core_fclasses_feats}
    core_fclasses_targets_map = {fclass.__name__: fclass for fclass in core_fclasses_targets}
    native_fclasses_map = {fclass.__name__: fclass for fclass in native_fclasses}

    # Create canonical featurizer list if "all" options are selected
    if "all_features" in featurize_with:
        featurize_with = list(core_fclasses_feats_map.keys())
    elif "all_targets" in featurize_with:
        featurize_with = list(core_fclasses_targets_map.keys())

    # Feature class names along with hyperparameters
    # These are all default
    fclass_names_w_params = [(fclass_name, None) for fclass_name in featurize_with]

    # Add featurizers with custom parameters to list of featurizers to apply
    for fstr in featurize_with_hyperparams:
        fdict = ast.literal_eval(fstr)
        if not isinstance(fdict, dict):
            raise TypeError(f"Could not parse input featurizer with parameters string {fdict}")
        if len(fdict) != 1:
            raise ValueError(f"Featurizer must be specified as sole root key of hyperparam dictionary: {fdict}")
        fclass_name_w_params = [(k, v) for k, v in fdict.items()][0]
        fclass_names_w_params.append(fclass_name_w_params)

    # Determine actual classes to apply by joining with external modules
    fclass_tuples = []
    for fclass_name, fclass_params in fclass_names_w_params:
        if fclass_name in native_fclasses_map:
            fclass = native_fclasses_map[fclass_name]
        else:
            # it is assumed it will be an external module
            if "." not in fclass_name:
                logging.critical(
                    f"'{fclass_name}' not recognized as BEEP native featurizer "
                    f"or importable module."
                )
                click.Context.exit(1)

            modname, _, clsname = fclass_name.rpartition('.')
            mod = importlib.import_module(modname)
            cls = getattr(mod, clsname)

            if not issubclass(cls, BEEPFeaturizer):
                logging.critical(f"Class {cls.__name__} is not a subclass of BEEPFeaturizer.")
                click.Context.exit(1)
            fclass = cls

        # check parameter arguments and update with full hyperparameter specifications
        hps = copy.deepcopy(fclass.DEFAULT_HYPERPARAMETERS)
        if fclass_params is not None:
            hps.update(fclass_params)
        fclass_tuples.append((fclass, hps))

    logger.info(f"Applying {len(fclass_tuples)} featurizers to each of {n_files} files")

    # ragged featurizers apply is ok

    status_json = {
        "op_type": "featurize",
        "feature_matrix": None,
        "files": {}
    }
    i = 0
    bfs = []
    for file in files:
        log_prefix = f"File {i + 1} of {n_files}"

        t0_file = time.time()
        logger.debug(f"Hashing file '{file}' to MD5")
        op_result = {
            "walltime": None,
            "processed_md5_chksum": md5sum(file),
            "featurizers": []
        }

        logger.debug(f"{log_prefix}: Loading processed run '{file}'.")
        structured_datapath = auto_load_processed(file)
        logger.debug(f"{log_prefix}: Loaded processed run '{file}' into memory.")

        for fclass, f_hyperparams in fclass_tuples:
            fclass_name = fclass.__name__
            op_subresult = {
                "featurizer_name": fclass_name,
                "hyperparameters": f_hyperparams,
                "output": None,
                "valid": False,
                "featurized": False,
                "walltime": None,
                "traceback": None,
                "subop_md5_chksum": None
            }

            t0 = time.time()
            try:

                f = fclass(
                    structured_datapath=structured_datapath,
                    hyperparameters=f_hyperparams
                )

                is_valid, reason = f.validate()

                if is_valid:
                    op_subresult["valid"] = True
                    logger.info(f"{log_prefix}: Featurizer {fclass_name} valid with "
                                f"params {f_hyperparams} for '{file}'")
                else:
                    raise BEEPFeaturizationError(reason)

                f.create_features()
                op_subresult["featurized"] = True
                logger.info(
                    f"{log_prefix}: Featurizer {fclass_name} applied with params {f_hyperparams} for '{file}'")
                bfs.append(f)

                if save_intermediates:
                    intermediate_filename = f"{fclass_name}-{os.path.basename(file)}"
                    output_path = os.path.join(output_dir, intermediate_filename)
                    dumpfn(f, output_path)
                    logger.info(
                        f"{log_prefix}: Featurizer {fclass_name} features for '{file}' written to '{output_path}'")
                    op_subresult["output"] = output_path
                    logger.debug(f"Hashing sub-operation output file '{output_path}' to MD5")
                    op_subresult["subop_md5_chksum"] = md5sum(output_path)

            except KeyboardInterrupt:
                logger.critical("Keyboard interrupt caught - exiting...")
                click.Context.exit(1)

            except BaseException:
                tbinfo = sys.exc_info()
                logger.error(
                    f"{log_prefix}: Failed/invalid: ({tbinfo[0].__name__}): {fclass.__name__}")
                op_subresult["traceback"] = traceback.format_exception(*tbinfo)

                if ctx.obj.halt_on_error:
                    raise

            t1 = time.time()
            op_subresult["walltime"] = t1 - t0

            op_result["featurizers"].append(op_subresult)

        t1_file = time.time()
        op_result["walltime"] = t1_file - t0_file
        status_json["files"][file] = op_result
        i += 1

    feature_matrix_status = {}
    try:
        feature_matrix = BEEPFeatureMatrix(bfs)
        default_filename = f"FeatureMatrix-{datetime.datetime.now().strftime('%Y-%d-%m_%H.%M.%S.%f')}.json.gz"

        # Output filename specification overrides output dir
        if output_filename:
            output_filename = os.path.abspath(output_filename)
        else:
            output_filename = os.path.join(ctx.obj.cwd, default_filename)

        dumpfn(feature_matrix, output_filename)
        logger.info(f"Feature matrix of size {feature_matrix.matrix.shape} s"
                    f"uccessfully created and saved to {output_filename}")
        feature_matrix_status["created"] = True
        feature_matrix_status["traceback"] = None
        feature_matrix_status["output"] = output_filename
    except BaseException:
        if ctx.obj.halt_on_error:
            raise

        tbinfo = sys.exc_info()
        logging.critical(f"Feature matrix could not be created: '{tbinfo[0].__name__}'!")
        feature_matrix_status["created"] = False
        feature_matrix_status["traceback"] = traceback.format_exception(*tbinfo)
        feature_matrix_status["output"] = None

    status_json["feature_matrix"] = feature_matrix_status

    # Generate a summary output

    logger.info("Featurization report:")

    all_succeeded, some_succeeded, none_succeeded = [], [], []
    for file, data in status_json["files"].items():
        feats_succeeded = []

        for fdata in data["featurizers"]:
            feats_succeeded.append(fdata["featurized"])

        n_success = sum(feats_succeeded)
        if n_success == len(fclass_tuples):
            all_succeeded.append((file, n_success))
        elif n_success == 0:
            none_succeeded.append((file, n_success))
        else:
            some_succeeded.append((file, n_success))

    logger.info(f"\tAll {len(fclass_tuples)} featurizers succeeded: {len(all_succeeded)}/{n_files}")
    if len(all_succeeded) > 0:
        for filename, _ in all_succeeded:
            logger.info(f"\t\t- {filename}")

    if len(fclass_tuples) > 1:
        logger.info(f"\tSome featurizers succeeded: {len(some_succeeded)}/{n_files}")
        if len(some_succeeded) > 0:
            for filename, n_success in some_succeeded:
                logger.info(f"\t\t- {filename}: {n_success}/{len(fclass_tuples)}")

    logger.info(f"\tNo featurizers succeeded or file failed: {len(none_succeeded)}/{n_files}")
    if len(none_succeeded) > 0:
        for filename, _ in none_succeeded:
            logger.info(f"\t\t- {filename}")

    logger.info(f"\tFeaturization matrix created: {status_json['feature_matrix']['created']}")

    status_json = add_metadata_to_status_json(status_json, ctx.obj.run_id, ctx.obj.tags)
    osj = ctx.obj.output_status_json
    if osj:
        dumpfn(status_json, osj)
        logger.info(f"Wrote status json file to {osj}")


@cli.command(
    help="Train a machine learning model using all available data "
         "and save it to file.",
    short_help="train a model with existing features"
)
@click.option(
    "--output-filename",
    "-o",
    required=False,
    type=CLICK_FILE,
    help="Filename (json) to write the BEEP linear model object to "
         "when training is finished."
)
@click.option(
    '--feature-matrix-file',
    '-fm',
    required=True,
    type=CLICK_FILE,
    help="Featurization matrix serialized to file, containing "
         "features (X) for learning. Featurization matrices can "
         "be generated by the beep featurize command."
)
@click.option(
    '--target-matrix-file',
    '-tm',
    required=True,
    type=CLICK_FILE,
    help="Featurization matrix serialized to file, containing "
         "targets (one y or more) for learning. Featurization "
         "matrices can be generated by the beep featurize command."
)
@click.option(
    '--targets',
    '-t',
    required=True,
    multiple=True,
    type=click.STRING,
    help="Target columns to as from target matrix file. Must all "
         "be present in the target matrix file. If more than 1 is "
         "specified (e.g., -t 'col1' -t 'col2'), multitask regression "
         "will be performed. Column names will be '<Feature Name>::"
         "<Featurizer Class Name>' if --homogenize-features is "
         "set. If not, column names include long parameter hashes "
         "which must be included in this argument option."
)
@click.option(
    "--model-name",
    "-m",
    required=True,
    type=click.STRING,
    help="Name of the regularized linear model to use. Current selection "
         f"includes {BEEPLinearModelExperiment.ALLOWED_MODELS}."
)
@click.option(
    "--train-on-frac-and-score",
    "-s",
    type=click.FLOAT,
    help="Do hyperparameter tuning on part (a training fraction) of the "
         "dataset and use that fitted model to predict on a testing "
         "fraction of the dataset. Specify the training fraction as "
         "a float 0-1."
)
@click.option(
    "--alpha-lower",
    "-al",
    type=click.FLOAT,
    help="Lower bound on the grid for the alpha hyperparameter "
         " which will be explored during hyperparameter tuning. "
         "Must be specified with --alpha-upper and --n-alphas."
)
@click.option(
    "--alpha-upper",
    "-au",
    type=click.FLOAT,
    help="Upper bound on the grid for the alpha hyperparameter "
         " which will be explored during hyperparameter tuning. "
         "Must be specified with --alpha-lower and --n-alphas."
)
@click.option(
    "--n-alphas",
    "-an",
    type=click.FLOAT,
    help="Number of linearly spaced alphas to explore during "
         "hyperparameter tuning. If not specified, sklearn defaults "
         "are used. Must be specified with --alpha-upper and "
         "--alpha-lower."
)
@click.option(
    "--train-feature-nan-thresh",
    type=click.FLOAT,
    help="Threshold to keep a feature in the training dataset, in "
         "fraction of samples which must not be nan from 0-1. "
         "0 = any feature having any nan is dropped, 1 = no "
         "features are dropped."
)
@click.option(
    "--train-sample-nan-thresh",
    type=click.FLOAT,
    help="Threshold to keep a sample from the training data, in "
         "fraction of features which must not be nan "
         "from 0-1. 0 = any sample having any "
         "nan feature is dropped, 1 = no samples are dropped."
)
@click.option(
    "--predict-sample-nan-thresh",
    type=click.FLOAT,
    help="Threshold to keep a sample from any prediction set, including "
         "those used internally, in fraction of features which must not"
         "be nan."
)
@click.option(
    "--drop-nan-training-targets",
    is_flag=True,
    default=True,
    help="Drop samples containing any nan targets. If False and "
         "the targets matrix has nan targets, the command will fail."
)
@click.option(
    "--impute-strategy",
    type=click.STRING,
    help="Type of imputation to use, 'median', 'mean', or 'none'."
)
@click.option(
    "--kfold",
    type=click.INT,
    help="Number of folds to use in k-fold hyperparameter tuning."
)
@click.option(
    "--max-iter",
    type=click.INT,
    help="Number of iterations during training to fit linear parameters."
)
@click.option(
    "--tol",
    type=click.FLOAT,
    help="Tolerance for optimization."
)
@click.option(
    "--l1-ratios",
    type=click.STRING,
    help="Comma separated l1 ratios to try in hyperparameter optimization."
         "For example, '0.1,0.5,0.7,0.9,0.95,1.0', and all values must be "
         "between 0-1."
)
@click.option(
    "--homogenize-features",
    type=click.BOOL,
    help="Shorten feature names to only include the featurizer name and "
         "(very short) feature name. For example, "
         "'capacity_0.8::TrajectoryFastCharge', where features normally "
         "have names including their (long) parameter hashes. To use the "
         "literal feature names, specify False."
)
@click.pass_context
def train(
        ctx,
        output_filename,
        feature_matrix_file,
        target_matrix_file,
        targets,
        model_name,
        train_on_frac_and_score,
        alpha_lower,
        alpha_upper,
        n_alphas,
        train_feature_nan_thresh,
        train_sample_nan_thresh,
        predict_sample_nan_thresh,
        drop_nan_training_targets,
        impute_strategy,
        kfold,
        max_iter,
        tol,
        l1_ratios,
        homogenize_features
):
    """
    Use a feature matrix and a target matrix to train a machine learning model.
    Automatically performs hyperparameter tuning of linear models and can
    evaluate test fractions automatically, returning error metrics in the
    status and logs.


    feature matrix json + target matrix json ---> trained model json


    Args:
        ctx (ContextPersister): A context-persisting object containing
            metadata and logging specifications for subcommands.
        output_filename (str): The output filename to save the trained model
            to. Must be json.
        feature_matrix_file (str): Path of the file to use as the learning
            features, X. Must be a serialized BEEPFeatureMatrix.
        target_matrix_file (str): Path of the file containing the target
            features, y or Y (if multiple). Must be a serialized
            BEEPFeatureMatrix.
        targets ([str]): One or more literal target columns in the target matrix to
            use as learning target or targets. Must be present in the target
            matrix. If homogenize_features is False, must be the full name
            of the feature, including the parameter hash.
        model_name (str): Model name to train. E.g., "elasticnet", OR "lasso",
            OR "ridge".
        train_on_frac_and_score (float): If set, this will train on a fraction of
            data (specified by the float) and then predict on the rest, returning
            error metrics in the test fraction in the status json and log.
        alpha_lower (float): Lower bound for alpha for linear model search.
        alpha_upper (float): Upper bounds for alpha for the linear model search.
        n_alphas (int): Number of alphas to search in the range [alpha_lower,
            alpha_upper].
        train_feature_nan_thresh (float): Threshold fraction of non-nan a feature
            column in the training set must possess in order to be retained, 0-1. 0=
            features are never dropped (may cause errors), 1= any feature containing
            any nan sample is dropped.
        train_sample_nan_thresh (float): Threshold fraction of non-nan features a sample
            row in the training set must possess in order to be retained, 0-1. 0=
            samples are never dropped (may cause errors), 1= any sample containing
            any nan feature is dropped.
        predict_sample_nan_thresh (float): Threshold fraction of non-nan features a sample
            row in the prediction set must possess in order to be retained, 0-1. 0=
            samples are never dropped (may cause errors). 1= any sample containing
            any nan feature is dropped.
        drop_nan_training_targets (bool): Drop any samples in the training set
            which do not have non-nan targets.
        impute_strategy (str): Select an impute strategy for samples in training
            and prediction which, even after dropping nan columns and samples
            according to thresholds, still have nans.
        kfold (int): Number of folds to perform cross-validation and hyperparameter
            optimization with.
        max_iter (int): Number of iterations to use for fitting linear models.
        tol (float): Tolerance for optimization.
        l1_ratios (str): Comma-separated l1 ratios to use in hyperparameter
            optimization for linear models.
        homogenize_features (bool): If True, shortens feature names to only
            include the featurizer and feature name. For example,
            'capacity_0.8::TrajectoryFastCharge'. Can be used for merging disparate
            features created with different hyperparameters, as long as the feature
            name and the featurizer name match. If False, keeps the parameter hash
            of the feature in the feature name. Set to False if using features
            generated by identical featurizer classes but different hyperparameters.
            If False, must specify the long feature name in the targets including
            the parameter hash.

    Returns:
        None

    """
    feature_matrix_file = os.path.abspath(feature_matrix_file)
    target_matrix_file = os.path.abspath(target_matrix_file)
    targets = list(targets)

    logger.info(
        f"Running training using files {feature_matrix_file} (features) "
        f"and {target_matrix_file} (targets)."
    )

    alpha_params = [alpha_lower, alpha_upper, n_alphas]
    if any(alpha_params) and not all(alpha_params):
        alphas = np.linspace(start=alpha_lower, stop=alpha_upper, num=n_alphas)
    else:
        alphas = None

    if l1_ratios is not None:
        l1_ratios = [float(l1r) for l1r in l1_ratios.strip().split(",")]

    additional_kwargs = {
        "train_feature_nan_thresh": train_feature_nan_thresh,
        "train_sample_nan_thresh": train_sample_nan_thresh,
        "predict_sample_nan_thresh": predict_sample_nan_thresh,
        "drop_nan_training_targets": drop_nan_training_targets,
        "impute_strategy": impute_strategy,
        "kfold": kfold,
        "max_iter": max_iter,
        "tol": tol,
        "l1_ratios": l1_ratios,
        "homogenize_features": homogenize_features,
        "alphas": alphas,
    }

    # only pass in arguments which will override the defaults of the lower class
    additional_kwargs = {k: v for k, v in additional_kwargs.items() if v is not None}

    logger.debug(f"Hashing file '{feature_matrix_file}' to MD5")
    logger.debug(f"Hashing file '{target_matrix_file}' to MD5")
    status_json = {
        "op_type": "train",
        "files": {
            "features": feature_matrix_file,
            "targets": target_matrix_file,
            "features_md5_chksum": md5sum(feature_matrix_file),
            "targets_md5_chksum": md5sum(target_matrix_file)
        },
        "model_results": {},
        "walltime": None,
        "trained_model": {
            "created": False,
            "output": None,
        },
        "traceback": None,
    }

    t0 = time.time()
    model, training_errors, test_errors = None, None, None
    try:
        bfm = BEEPFeatureMatrix.from_json_file(feature_matrix_file)
        btm = BEEPFeatureMatrix.from_json_file(target_matrix_file)

        blme = BEEPLinearModelExperiment(
            feature_matrix=bfm,
            target_matrix=btm,
            targets=targets,
            model_name=model_name,
            **additional_kwargs
        )

        if train_on_frac_and_score is not None:
            logger.info("Beginning training and scoring on test set.")
            model, training_errors, test_errors = blme.train_and_score(train_and_val_frac=train_on_frac_and_score)
            status_json["model_results"]["training_error"] = training_errors
            status_json["model_results"]["test_error"] = test_errors
            status_json["model_results"]["test_fraction"] = train_on_frac_and_score
            status_json["model_results"]["optimal_hyperparameters"] = blme.optimal_hyperparameters
        else:
            logger.info("Beginning training on all available data")
            model, training_errors = blme.train()
            test_errors = None
    except KeyboardInterrupt:
        logging.critical("Keyboard interrupt caught - exiting...")
        click.Context.exit(1)
    except BaseException:
        tbinfo = sys.exc_info()
        tbfmt = traceback.format_exception(*tbinfo)
        logger.error(f"Model training failed: ({tbinfo[0].__name__})")
        status_json["traceback"] = tbfmt

        if ctx.obj.halt_on_error:
            raise

    t1 = time.time()
    status_json["walltime"] = t1 - t0

    if model:
        logger.info(f"Model {model} trained, finding optimal hyperparamters {blme.optimal_hyperparameters}.")
        logger.info(f"Training error summary: {training_errors}")
        if test_errors:
            logger.info(f"Testing error summary: {test_errors}")

        default_filename = f"LinearModelExperiment-{datetime.datetime.now().strftime('%Y-%d-%m_%H.%M.%S.%f')}.json.gz"
        if output_filename:
            output_filename = os.path.abspath(output_filename)
        else:
            output_filename = os.path.join(ctx.obj.cwd, default_filename)
        dumpfn(blme, output_filename)
        logger.info(f"Wrote model {model} to path: {output_filename}")
        status_json["trained_model"]["created"] = True
        status_json["trained_model"]["output"] = output_filename

    status_json = add_metadata_to_status_json(status_json, ctx.obj.run_id, ctx.obj.tags)
    osj = ctx.obj.output_status_json
    if osj:
        dumpfn(status_json, osj)
        logger.info(f"Wrote status json file to {osj}")


@cli.command(
    help="Run a previously trained model to predict degradation targets."
         "The MODEL_FILE passed should be an output of 'beep train' or a"
         "serialized BEEPLinearModelExperiment object.",
    short_help="use a trained model to predict degradation"
)
@click.argument(
    "model_file",
    nargs=1,
    type=CLICK_FILE
)
@click.option(
    "--feature-matrix-file",
    "-fm",
    required=True,
    multiple=False,
    help="Feature matrix to use as input to the model. Predictions are based"
         "on these features."
)
@click.option(
    "--output-filename",
    "-o",
    required=False,
    type=CLICK_FILE,
    help="Filename (json) to write the final predicted dataframe to."
)
@click.option(
    "--predict-sample-nan-thresh",
    type=click.FLOAT,
    help="Threshold to keep a sample from any prediction set."
)
@click.pass_context
def predict(
        ctx,
        model_file,
        feature_matrix_file,
        output_filename,
        predict_sample_nan_thresh,

):
    """
    Predict degradation characteristics using a previously trained model.


    trained model json + new feature matrix json ---> predictions json


    Args:
        ctx (ContextPersister): A context-persisting object containing
            metadata and logging specifications for subcommands.
        model_file (str): The json or json.gz of a trained model file. Must
            be a BEEPLinearModelExperiment class serialized to disk.
        feature_matrix_file (str): Path to a feature matrix file to use
            as input features for inference. These features will be used to
            predict the degradation targets the previously trained model
            was fit on.
        output_filename (str): Path to output predictions to. Will be saved in
            dataframe format as a json.
        predict_sample_nan_thresh (float): Threshold fraction of features a sample
            row in the prediction set must possess in order to be dropped, 0-1. 0=
            samples are never dropped (may cause errors). 1= any sample containing
            any nan feature is dropped. Overrides any predict sample nan threshold
            set during model training.

    Returns:
        None

    """
    if output_filename and not output_filename.endswith(".json"):
        raise ValueError("--output-filename must end with '.json'.")

    default_filename = f"PredictedDegradationDF-" \
                       f"{datetime.datetime.now().strftime('%Y-%d-%m_%H.%M.%S.%f')}" \
                       f".json"
    if output_filename:
        output_filename = os.path.abspath(output_filename)
    else:
        output_filename = os.path.join(ctx.obj.cwd, default_filename)

    model_file = os.path.abspath(model_file)
    feature_matrix_file = os.path.abspath(feature_matrix_file)

    logger.debug(f"Hashing file '{model_file}' to MD5")
    logger.debug(f"Hashing file '{feature_matrix_file}' to MD5")
    status_json = {
        "op_type": "predict",
        "files": {
            "model": model_file,
            "predict_on_features": feature_matrix_file,
            "model_md5_chksum": md5sum(model_file),
            "predict_on_features_md5_chksum": md5sum(feature_matrix_file)
        },
        "walltime": None,
        "predictions": {
            "created": False,
            "output": None
        },
        "traceback": None,
    }

    t0 = time.time()
    predicted = None
    try:
        logger.debug(
            f"Loading linear model from file on disk at path: {model_file}")

        bfm = BEEPFeatureMatrix.from_json_file(feature_matrix_file)
        model = BEEPLinearModelExperiment.from_json_file(model_file)

        # Optionally override the previously set sample nan thresh
        # for prediction dataframes
        if predict_sample_nan_thresh:
            model.predict_sample_nan_thresh = predict_sample_nan_thresh

        logging.debug(
            f"Model instantiated ok, predicting on {bfm.matrix.shape[0]}"
            f" inference samples."
        )

        predicted, dropped_ix = model.predict(bfm)
        logging.info(
            f"Successfully predicted {predicted.shape[0]} samples of "
            f"{bfm.matrix.shape[0]} original samples."
        )

    except KeyboardInterrupt:
        logging.critical("Keyboard interrupt caught - exiting...")
        click.Context.exit(1)
    except BaseException:
        tbinfo = sys.exc_info()
        tbfmt = traceback.format_exception(*tbinfo)
        logger.error(f"Model inference failed: ({tbinfo[0].__name__})")
        status_json["traceback"] = tbfmt

        if ctx.obj.halt_on_error:
            raise

    t1 = time.time()
    status_json["walltime"] = t1 - t0

    if predicted is not None:
        predicted.to_json(output_filename)
        status_json["predictions"]["created"] = True
        status_json["predictions"]["output"] = output_filename
        logger.info(f"Successfully wrote predicted dataframe to {output_filename}.")

    status_json = add_metadata_to_status_json(status_json, ctx.obj.run_id, ctx.obj.tags)
    osj = ctx.obj.output_status_json
    if osj:
        dumpfn(status_json, osj)
        logger.info(f"Wrote status json file to {osj}")


@cli.command(
    help="Generate protocol for battery cyclers from a csv file input.",
    short_help="generate protocol for new cycler experiments"
)
@click.argument(
    "csv_file",
    nargs=1,
    type=CLICK_FILE
)
@click.option(
    "--output-dir",
    "-d",
    help="Directory to output files to. At least three subdirs will be created in this directory"
         "in order to organize the generated protocol files."
)
@click.pass_context
def protocol(
        ctx,
        csv_file,
        output_dir
):
    """
    Generate cycler protocol files using a templating csv and set of template files.


    WARNING: This CLI function is still under migration and may not work as intended.



    protocol generating csv ---> protocol files


    Args:
        ctx (ContextPersister): A context-persisting object containing
            metadata and logging specifications for subcommands.
        csv_file (str): Path to the protocol generating csv.
        output_dir (str): Path to the output directory where protocol files
            should be kept. Subdirectories are automatically created for
            various protocol file types.

    Returns:
        None

    """
    click.secho("Protocol generation has yet to be migrated entirely to the new CLI. May be unstable.", fg="red")

    cwd = ctx.obj.cwd

    csv_file = os.path.abspath(csv_file)
    if not os.path.exists(csv_file):
        raise FileNotFoundError(f"Input file {csv_file} does not exist")

    output_dir = os.path.abspath(output_dir) if output_dir else cwd
    if not os.path.exists(output_dir):
        raise FileNotFoundError(f"Output dir {output_dir} does not exist.")

    logger.debug(f"Generating protocol based on '{csv_file}' into '{output_dir}'.")

    status_json = {
        "op_type": "protocol",
        "files": {
            "csv": csv_file,
            "csv_md5_chksum": md5sum(csv_file)
        },
        "walltime": None,
        "protocol": {},
        "traceback": None,
    }

    t0 = time.time()
    output_files, failures, result, message = [], [], "", {}
    try:
        results = generate_protocol_files_from_csv(csv_file, output_directory=output_dir)
        output_files, file_generation_failures, result, message = results

        if result == "error" or failures:
            msg = message["comment"]
            err = message["error"]
            raise ProtocolException(f"'{err}: {msg}'")
        else:
            logger.info(f"Created {len(output_files)} protocol files in `{output_dir}`.")
    except BaseException:
        tbinfo = sys.exc_info()
        tbfmt = traceback.format_exception(*tbinfo)
        logger.error(f"Protocol generation failed with : ({tbinfo[0].__name__})")
        status_json["traceback"] = tbfmt
    t1 = time.time()

    for of in output_files:
        status_json["protocol"][of] = {"generated": True}

    for f in failures:
        status_json["protocol"][f] = {"generated": False}

    status_json["walltime"] = t1 - t0
    status_json = add_metadata_to_status_json(status_json, ctx.obj.run_id, ctx.obj.tags)
    osj = ctx.obj.output_status_json
    if osj:
        dumpfn(status_json, osj)
        logger.info(f"Wrote status json file to {osj}")


@cli.command(
    help="View BEEP files for debugging and analysis.",
    short_help="viewer for beep files, help debug"
)
@click.argument(
    "file",
    nargs=1,
    type=CLICK_FILE
)
def inspect(file):
    """
    Inspect any BEEP object serialized to disk from its filename.

    Args:
        file (str): Path of the file to inspect. Should be json or json.gz.

    Returns:
        None

    """
    file = os.path.abspath(file)
    try:
        o = loadfn(file)

        # To weed out pesky datapaths from old serializations
        o.as_dict()
    except (AttributeError, ValueError, KeyError):

        # Maybe it is a structured file?
        if file.endswith(".json") or file.endswith(".json.gz"):
            try:
                o = auto_load_processed(file)
            except BaseException:
                raise ValueError(f"File format of json file {file} not "
                                 f"recognized with beep version {__version__}")

        # Or it could be a raw file?
        else:
            try:
                o = auto_load(file)
                logger.debug(f"Loaded potential raw file {file} as Datapath.")
            except BaseException:
                raise ValueError(f"File format of {file} not loadable with beep.")

    logger.info(f"Loaded {file} as type {type(o)}.")

    if isinstance(o, BEEPDatapath):
        print(f"\n\nBEEP Datapath: {file}\n\n")

        print(f"\nSemiunique id: '{o.semiunique_id}'")

        print("\nFile paths")
        pprint.pprint(o.paths)

        print("\nFile metadata:")
        pprint.pprint(o.metadata.raw)

        print(f"\nValidation schema: {o.schema}")

        print("\nStructuring parameters:")
        pprint.pprint(o.structuring_parameters)

        print("\nStructured attributes:")

        for attr in [
            "structured_summary",
            "structured_data",
            "diagnostic_data",
            "diagnostic_summary",
            "raw_data",
        ]:

            print(f"\n{attr}:")
            a = getattr(o, attr)
            if isinstance(a, pd.DataFrame):
                print(a)
                print(a.info())
            else:
                print("\tNo object.")

    elif isinstance(o, BEEPFeaturizer):
        print("\n\nBEEP Featurizer: {file}\n\n")

        print("\nFile paths:")
        pprint.pprint(o.paths)

        print(f"\nLinked datapath semiunique id: {o.linked_semiunique_id}")

        print("\nHyperparameters:")
        pprint.pprint(o.hyperparameters)

        print("\nMetadata:")
        pprint.pprint(o.metadata)

        print("\nFeatures:")
        print(o.features)

    elif isinstance(o, BEEPFeatureMatrix):
        print(f"\n\nBEEP Feature Matrix: {file}\n\n")

        print("\nFeaturizers:")
        for f in o.featurizers:
            d_display = f.as_dict()
            d_display.pop("features")
            print(f"\n\tFeaturizer {f.__class__.__module__} {f.__class__.__name__}")
            d_formatted = pprint.pformat(d_display)
            d_formatted = "\t" + d_formatted.replace("\n", "\n\t")
            print(d_formatted)

        print("\nMatrix:")
        print(o.matrix)
        print(o.matrix.info())

    elif isinstance(o, BEEPLinearModelExperiment):
        d = o.as_dict()

        print(f"\n\nBEEP Linear Model Experiment: {file}\n\n")

        print(f"\nTargets: {o.targets}")

        print(f"\nModel name: {o.model_name}")

        print(f"\nImpute strategy: {o.impute_strategy}")

        print(f"\nHomogenize features: {o.homogenize_features}")

        print("\nNaN Thresholds:")
        for k in [
            "train_feature_drop_nan_thresh",
            "train_sample_drop_nan_thresh",
            "predict_sample_nan_thresh"
        ]:
            print(f"\t-{k}: {d[k]}")

        print("\nModel parameters:")
        for p, v in d["model_sklearn"].items():
            print(f"\t- {p}: {v}")

        print("\nMatrices:")
        for attr in ["feature_matrix", "target_matrix"]:
            a = getattr(o, attr).matrix

            print(f"\n{attr}")
            print(a)
            print(a.info())

    else:
        pprint.pprint(o.as_dict())
        logger.warning(f"Could not identify object {file}! Dictionary representation printed.")



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Parsing and conversion of maccor procedure files to arbin schedule files"""

import os
import re
from datetime import datetime
from copy import deepcopy
from beep.protocol import PROTOCOL_SCHEMA_DIR
from collections import OrderedDict
from monty.serialization import loadfn
from beep.protocol.arbin import Schedule

TEST_DIR = os.path.dirname(__file__)
TEST_FILE_DIR = os.path.join(TEST_DIR, "test_files")


class ProcedureToSchedule:
    """
    This class is a set of methods to convert from a maccor procedure file to
    an arbin schedule file. This is essentially a translation between two
    different system languages. Since the two systems are not equivalent,
    its not possible to perform an exact translation and some control methods
    cannot be supported.

    Args:
        procedure_dict_steps (OrderedDict): A dictionary containing each of
            the steps in the procedure

    """

    def __init__(self, procedure_dict_steps):
        self.procedure_dict_steps = procedure_dict_steps

    def create_sdu(self, sdu_input_name, sdu_output_name,
                   current_range='Range1',
                   global_v_range=[2.5, 4.5],
                   global_temp_range=[-100, 100],
                   global_current_range=[-30, 30]):
        """
        Highest level function in the class. Takes a schedule file and replaces
        all of the steps with steps from the procedure file. Then writes the
        resulting schedule file to the output file.

        Args:
            sdu_input_name (str): the full path of the schedule file to use as a
                shell for the steps
            sdu_output_name (str): the full path of the schedule file to output
            current_range (str): Current 30A, 5A, 500mA, 20mA  ('Range1',2,3,4, respectively, also Parallel-High is
            an option)
            global_v_range (list): Global safety range for voltage in volts [min, max]
            global_temp_range (list): Global safety range for temperature in Celcius [min, max]
            global_current_range (list): Global safety range for current in Amps [min, max]

        """
        schedule = Schedule.from_file(sdu_input_name)
        # sdu_dict = Schedule.from_file(sdu_input_name)

        keys = list(schedule['Schedule'].keys())
        safety_data = []
        safety_key = []
        for key in keys:
            if bool(re.match('Step[0-9]+', key)):
                del schedule['Schedule'][key]
            elif bool(re.match('UserDefineSafety[0-9]+', key)):
                safety_data.append(deepcopy(schedule['Schedule'][key]))
                safety_key.append(key)
                del schedule['Schedule'][key]

        step_name_list, step_flow_ctrl = self.create_metadata()
        schedule.set("Schedule.m_uStepNum", str(len(step_name_list)))
        # Set global safety limits
        schedule.set("Schedule.m_VoltageSafetyScope",
                     "{:.6f}".format(global_v_range[0]) + "^" + "{:.6f}".format(global_v_range[1]))
        schedule.set("Schedule.m_AuxTempSafetyScope",
                     "{:.6f}".format(global_temp_range[0]) + '^' + "{:.6f}".format(global_temp_range[1]))
        schedule.set("Schedule.m_CurrentSafetyScope",
                     "{:.6f}".format(global_current_range[0]) + '^' + "{:.6f}".format(global_current_range[1]))
        if global_temp_range != [-100, 100]:
            schedule.set("Schedule.m_AuxSafetyEnabled",
                         '0^0 ; 1^1 ; 2^0 ; 3^0 ; 4^0 ; 5^0 ; 6^0 ; 7^0 ; 8^0 ; 9^0 ; 10^0 ; 11^0')
        for step_index, step in enumerate(self.procedure_dict_steps):

            step_arbin = self.compile_to_arbin(
                self.procedure_dict_steps[step_index],
                step_index,
                step_name_list,
                step_flow_ctrl,
                current_range=current_range
            )
            key = "Step{}".format(step_index)

            schedule.set("Schedule.{}".format(key), step_arbin)
        for indx, safety in enumerate(safety_data):
            schedule.set("Schedule.{}".format(safety_key[indx]), safety)
        schedule.to_file(sdu_output_name)

    def create_metadata(self):
        """
        Creates the necessary information so that flow control for the steps can be
        set up correctly.

        Returns:
            list: unique strings consisting of the step number and note
            dict: key value pairs for the loop control steps indicating
                which step the Loop step should GOTO when the end condition
                is not matched

        """
        step_name_list = []
        step_flow_ctrl = {}
        for indx, step in enumerate(self.procedure_dict_steps):
            step_name_list.append(str(indx + 1) + "-" + str(step["StepNote"]))
            if "Loop" in step["StepType"]:
                loop_counter = int(re.search(r"\d+", step["StepType"]).group())
                i = 1
                while (indx - i) > 1:
                    if self.procedure_dict_steps[abs(indx - i)][
                        "StepType"
                    ] == "Do {}".format(loop_counter):
                        do_index = abs(indx - i)
                        break
                    i = i + 1
                step_flow_ctrl.update({indx: step_name_list[do_index + 1]})
        return step_name_list, step_flow_ctrl

    def compile_to_arbin(
        self,
        step_abs,
        step_index,
        step_name_list,
        step_flow_ctrl,
        current_range="Parallel-High",
    ):
        """
        Takes a step from a maccor procedure file and converts it to an equivalent
        step for an arbin schedule file. Not all control modes are supported. Compatible
        modes are constant current, constant voltage, CCCV. Future support for power,
        resistance modes and complex modes such as formulas and waveforms is planned.

        Flow control for the steps is implemented by using the TC counters of the arbin
        schedule. Since only 4 counters are available, only 4 layer of Loops are possible.
        Do steps are converted to steps that reset the respective TC counter and the Loop
        steps are converted to steps that go to the step immediately after the Do step until
        the loop condition is met, while advancing the TC counter every time the Loop step is
        passed.

        Arbin control values decoded: Arbin is using a 20-bit value for control and masking
        off specific portions for each of the different actions (reset, increment, decrement)
        and using different bits for each of the different counters
        =====================================================================================================
        Step Number	Name	    CtrlValue	Ext1	Ext2    Reset Increment
        3	Loop (CI 0)	                0	   1	  0				    CI			                    00001
        7	First Loop (Reset)	        0	   2	  0				    T1			                    00010
        8	Loop (CI 1)	                0	   1	  0				    CI			                    00001
        16	Loop 2 ( HPPC)	            0	   2	  0				    T1			                    00010
        17	Loop (CI 2)	                0	   1	  0				    CI			                    00001
        21	Loop (CI 3)     	        0	   1	  0				    CI			                    00001
        25	Loop (CI 4)	                0	   1	  0				    CI			                    00001
        29	Loop (CI 5)	                0	   1	  0				    CI			                    00001
        35	Loop 3 (Cycling 30)	    65536	   13	  0			T1	    T2		00010 00000 00000 00000	01101
        36	Loop 4 (Cycling 100)	    0	   16	  0				    T4			                    10000
        37	Loop 5 (always true)	524288	   1	  0			T4	    CI		10000 00000 00000 00000	00001
        15	Loop                       15      1      0         PVs             00000 00000 00000 01111 00001

        Args:
            step_abs (OrderedDict): A ordered dict of the maccor step to be converted
            step_index (int): The index of the step to be converted
            step_name_list (list): A list of the step labels to be used in the arbin
                schedule file
            step_flow_ctrl (dict): A dictionary of the loop steps as keys and the
                corresponding steps to go to after the
            current_range (str): The current range to use for the step, values can
            be 'Range1', 'Range2', 'Range3', 'Range4' and 'Parallel-High' depending on the
            cycler being used. Range1 is the largest normal current range. The values for the
            ranges are 30A, 5A, 500mA, 20mA

        Returns:
            OrderedDict: The arbin step resulting from the conversion of the
                procedure step
        """

        ARBIN_SCHEMA = loadfn(
            os.path.join(PROTOCOL_SCHEMA_DIR, "arbin_schedule_schema.yaml")
        )
        blank_step = OrderedDict(ARBIN_SCHEMA["step_blank_body"])

        blank_step["m_szLabel"] = str(step_index + 1) + "-" + str(step_abs["StepNote"])
        blank_step["m_szCurrentRange"] = current_range

        # Current control mode with currents measured in Amps
        if step_abs["StepMode"] == "Current" and "C" not in step_abs["StepValue"]:
            if step_abs["Limits"] is not None:
                blank_step["m_szStepCtrlType"] = "CCCV"
                if step_abs["StepType"] == "Charge":
                    blank_step["m_szCtrlValue"] = step_abs["StepValue"]
                    blank_step["m_szExtCtrlValue1"] = step_abs["Limits"]["Voltage"]
                    blank_step["m_szExtCtrlValue2"] = "0"
                elif step_abs["StepType"] == "Dischrge":
                    blank_step["m_szCtrlValue"] = "-" + step_abs["StepValue"]
                    blank_step["m_szExtCtrlValue1"] = step_abs["Limits"]["Voltage"]
                    blank_step["m_szExtCtrlValue2"] = "0"
            elif step_abs["Limits"] is None:
                blank_step["m_szStepCtrlType"] = "Current(A)"
                if step_abs["StepType"] == "Charge":
                    blank_step["m_szCtrlValue"] = step_abs["StepValue"]
                elif step_abs["StepType"] == "Dischrge":
                    blank_step["m_szCtrlValue"] = "-" + step_abs["StepValue"]
            else:
                raise ValueError("Unable to set m_szStepCtrlType for current")

        # Current control mode currents measured in C-rate
        elif step_abs["StepMode"] == "Current" and "C" in step_abs["StepValue"]:
            if step_abs["Limits"] is not None:
                blank_step["m_szStepCtrlType"] = "CCCV"
                if step_abs["StepType"] == "Charge":
                    blank_step["m_szCtrlValue"] = step_abs["StepValue"].replace("C", "")
                    blank_step["m_szExtCtrlValue1"] = step_abs["Limits"]["Voltage"]
                    blank_step["m_szExtCtrlValue2"] = "0"
                elif step_abs["StepType"] == "Dischrge":
                    blank_step["m_szCtrlValue"] = "-" + step_abs["StepValue"].replace(
                        "C", ""
                    )
            elif step_abs["Limits"] is None:
                blank_step["m_szStepCtrlType"] = "C-Rate"
                if step_abs["StepType"] == "Charge":
                    blank_step["m_szCtrlValue"] = step_abs["StepValue"].replace("C", "")
                elif step_abs["StepType"] == "Dischrge":
                    blank_step["m_szCtrlValue"] = "-" + step_abs["StepValue"].replace(
                        "C", ""
                    )
            else:
                raise ValueError("Unable to set m_szStepCtrlType for current")

        # Voltage control mode and current limit measured in Amps
        elif step_abs["StepMode"] == "Voltage":
            if (
                step_abs["Limits"] is not None
                and "C" not in step_abs["Limits"]["Current"]
            ):
                blank_step["m_szStepCtrlType"] = "CCCV"
                if step_abs["StepType"] == "Charge":
                    blank_step["m_szCtrlValue"] = step_abs["Limits"]["Current"]
                    blank_step["m_szExtCtrlValue1"] = step_abs["StepValue"]
                    blank_step["m_szExtCtrlValue2"] = "0"
                elif step_abs["StepType"] == "Dischrge":
                    blank_step["m_szCtrlValue"] = "-" + step_abs["Limits"]["Current"]
                    blank_step["m_szExtCtrlValue1"] = step_abs["StepValue"]
                    blank_step["m_szExtCtrlValue2"] = "0"
            elif step_abs["Limits"] is None:
                blank_step["m_szStepCtrlType"] = "Voltage(V)"
                if step_abs["StepType"] == "Charge":
                    blank_step["m_szCtrlValue"] = step_abs["StepValue"]
                elif step_abs["StepType"] == "Dischrge":
                    blank_step["m_szCtrlValue"] = step_abs["StepValue"]
            else:
                raise ValueError("Unable to set m_szStepCtrlType for voltage")

        # Rest control mode
        elif step_abs["StepMode"] is None and step_abs["StepType"] == "Rest":
            if step_abs["Limits"] is None:
                blank_step["m_szStepCtrlType"] = "Rest"
            else:
                raise ValueError("Unable to set m_szStepCtrlType for voltage")

        # Flow control steps
        elif step_abs["StepMode"] is None and step_abs["StepType"] in [
            "Loop 1",
            "Do 1",
            "Loop 2",
            "Do 2",
            "AdvCycle",
            "End",
        ]:
            if step_abs["StepType"] == "AdvCycle":
                blank_step["m_szStepCtrlType"] = "Set Variable(s)"
                blank_step["m_szCtrlValue"] = "15"
                blank_step["m_szExtCtrlValue1"] = "1"
                blank_step["m_szExtCtrlValue2"] = "0"
            elif "Loop" in step_abs["StepType"]:
                loop_counter = int(re.search(r"\d+", step_abs["StepType"]).group())
                blank_step["m_szStepCtrlType"] = "Set Variable(s)"
                blank_step["m_szCtrlValue"] = "0"
                blank_step["m_szExtCtrlValue1"] = str(2 ** loop_counter)
                blank_step["m_szExtCtrlValue2"] = "0"
                assert isinstance(step_abs["Ends"]["EndEntry"], OrderedDict)
                loop_addendum = OrderedDict(
                    [
                        ("EndType", "Loop Addendum"),
                        ("Oper", "< "),
                        ("Step", step_flow_ctrl[step_index].split("-")[0]),
                        ("Value", step_abs["Ends"]["EndEntry"]["Value"]),
                    ]
                )
                step_abs["Ends"]["EndEntry"] = [
                    loop_addendum,
                    step_abs["Ends"]["EndEntry"],
                ]

            elif "Do" in step_abs["StepType"]:
                loop_counter = int(re.search(r"\d+", step_abs["StepType"]).group())
                blank_step["m_szStepCtrlType"] = "Set Variable(s)"
                blank_step["m_szCtrlValue"] = str(2 ** (loop_counter + 15))
                blank_step["m_szExtCtrlValue1"] = "0"
                blank_step["m_szExtCtrlValue2"] = "0"

            else:
                blank_step["m_szStepCtrlType"] = "Rest"
        else:
            raise ValueError("Unable to set StepMode for Flow control step")

        step_type = step_abs["StepType"]

        # Ends
        if step_abs["Ends"] is not None:
            if isinstance(step_abs["Ends"]["EndEntry"], OrderedDict):
                blank_step["m_uLimitNum"] = 1
                end = step_abs["Ends"]["EndEntry"]
                end_index = 0
                limit_key = "Limit{}".format(str(end_index))
                blank_step[limit_key] = OrderedDict(
                    self.convert_end_to_limit(
                        blank_step,
                        end,
                        step_index,
                        step_name_list,
                        step_type,
                        step_flow_ctrl,
                    )
                )
            elif isinstance(step_abs["Ends"]["EndEntry"], list):
                blank_step["m_uLimitNum"] = len(step_abs["Ends"]["EndEntry"])
                for end_index, end in enumerate(step_abs["Ends"]["EndEntry"]):
                    limit_key = "Limit{}".format(str(end_index))
                    blank_step[limit_key] = OrderedDict(
                        self.convert_end_to_limit(
                            blank_step,
                            end,
                            step_index,
                            step_name_list,
                            step_type,
                            step_flow_ctrl,
                        )
                    )
        elif step_abs["Ends"] is None:
            blank_step["m_uLimitNum"] = 1
            end_index = 0
            limit_key = "Limit{}".format(str(end_index))
            blank_step[limit_key] = self.add_blank_limit()

        # Reports
        if step_abs["Reports"] is not None:
            if isinstance(step_abs["Reports"]["ReportEntry"], OrderedDict):
                blank_step["m_uLimitNum"] = blank_step["m_uLimitNum"] + 1
                report = step_abs["Reports"]["ReportEntry"]
                report_index = 0
                limit_start = len(step_abs["Ends"]["EndEntry"])
                limit_key = "Limit{}".format(str(report_index + limit_start))
                blank_step[limit_key] = OrderedDict(
                    self.convert_report_to_logging_limit(report)
                )
            elif isinstance(step_abs["Ends"]["EndEntry"], list):
                blank_step["m_uLimitNum"] = blank_step["m_uLimitNum"] + len(
                    step_abs["Reports"]["ReportEntry"]
                )
                for report_index, report in enumerate(
                    step_abs["Reports"]["ReportEntry"]
                ):
                    limit_start = len(step_abs["Ends"]["EndEntry"])
                    limit_key = "Limit{}".format(str(report_index + limit_start))
                    blank_step[limit_key] = OrderedDict(
                        self.convert_report_to_logging_limit(report)
                    )

        blank_step["m_uLimitNum"] = str(blank_step["m_uLimitNum"])

        return blank_step

    def add_blank_limit(self):
        """
        Minor helper function to add a limit that immediately advances to the next step.

        Returns:
            dict: blank limit that advances to the next step immediately

        """
        ARBIN_SCHEMA = loadfn(
            os.path.join(PROTOCOL_SCHEMA_DIR, "arbin_schedule_schema.yaml")
        )
        limit = ARBIN_SCHEMA["step_blank_limit"]
        limit["m_bStepLimit"] = "1"
        limit["m_bLogDataLimit"] = "0"
        limit["m_szGotoStep"] = "Next Step"
        limit["Equation0_szLeft"] = "PV_CHAN_Step_Time"
        limit["Equation0_szCompareSign"] = ">"
        limit["Equation0_szRight"] = "0"
        return limit

    def convert_end_to_limit(
        self, blank_step, end, step_index, step_name_list, step_type, step_flow_ctrl
    ):
        """
        Takes a normal ending condition from a maccor procedure and converts it to an equivalent
        limit for an arbin schedule file.

        Args:
            blank_step (OrderedDict): the arbin step that is being populated
            end (OrderedDict): the ending condition from the maccor to convert to arbin limit
            step_index (int): the index of the current step being converted
            step_name_list (list): the list of labels for the steps
            step_type (str): the type of step being converted so that the limit can be set
                appropriately
            step_flow_ctrl (dict): a dictionary of the loop steps (keys) and the goto
                steps (values)

        Returns:
            dict: the converted limit

        """
        ARBIN_SCHEMA = loadfn(
            os.path.join(PROTOCOL_SCHEMA_DIR, "arbin_schedule_schema.yaml")
        )
        limit = ARBIN_SCHEMA["step_blank_limit"]
        limit["m_bStepLimit"] = "1"
        limit["m_bLogDataLimit"] = "1"

        if end["Step"] == str(int(step_index) + 2).zfill(3):
            limit["m_szGotoStep"] = "Next Step"
        else:
            limit["m_szGotoStep"] = step_name_list[int(end["Step"]) - 1]

        if end["EndType"] == "Voltage":
            limit["Equation0_szLeft"] = "PV_CHAN_Voltage"
            limit["Equation0_szCompareSign"] = end["Oper"].replace(" ", "")
            limit["Equation0_szRight"] = end["Value"]
        elif end["EndType"] == "Current" and blank_step["m_szStepCtrlType"] == "CCCV":
            limit["Equation0_szLeft"] = "PV_CHAN_CV_Stage_Current"
            if step_type == "Charge":
                limit["Equation0_szRight"] = end["Value"]
                limit["Equation0_szCompareSign"] = end["Oper"].replace(" ", "")
            elif step_type == "Dischrge":
                limit["Equation0_szRight"] = "-" + end["Value"]
                limit["Equation0_szCompareSign"] = (
                    end["Oper"].replace(" ", "").replace("<", ">")
                )
            else:
                raise ValueError(
                    "Unable to convert end to limit for EndType:{} and Ctrl:{}".format(
                        end["EndType"], blank_step["m_szStepCtrlType"]
                    )
                )
        elif (
            end["EndType"] == "Current"
            and blank_step["m_szStepCtrlType"] == "Voltage(V)"
        ):
            limit["Equation0_szLeft"] = "PV_CHAN_Current"
            if step_type == "Charge":
                limit["Equation0_szRight"] = end["Value"]
                limit["Equation0_szCompareSign"] = end["Oper"].replace(" ", "")
            elif step_type == "Dischrge":
                limit["Equation0_szRight"] = "-" + end["Value"]
                limit["Equation0_szCompareSign"] = (
                    end["Oper"].replace(" ", "").replace("<", ">")
                )
            else:
                raise ValueError(
                    "Unable to convert end to limit for EndType:{} and Ctrl:{}".format(
                        end["EndType"], blank_step["m_szStepCtrlType"]
                    )
                )
        elif end["EndType"] == "StepTime":
            limit["Equation0_szLeft"] = "PV_CHAN_Step_Time"
            limit["Equation0_szCompareSign"] = ">"
            if "." in end["Value"]:
                nofrag, frag = end["Value"].split(".")
                frag = frag[:6]  # truncate to microseconds
                frag += (6 - len(frag)) * "0"  # add 0s
                elapsed = datetime.strptime(
                    nofrag.replace("::", "00:00:0"), "%H:%M:%S"
                ).replace(microsecond=int(frag)) - datetime.strptime(
                    "00:00:00", "%H:%M:%S"
                )
            else:
                elapsed = datetime.strptime(
                    end["Value"].replace("::", "00:00:0"), "%H:%M:%S"
                ) - datetime.strptime("00:00:00", "%H:%M:%S")
            limit["Equation0_szRight"] = str(elapsed.total_seconds())
        elif end["EndType"] == "Loop Cnt":
            loop_counter = int(re.search(r"\d+", step_type).group())
            limit["Equation0_szLeft"] = "TC_Counter{}".format(loop_counter)
            limit["Equation0_szCompareSign"] = end["Oper"].replace(" ", "")
            limit["Equation0_szRight"] = end["Value"]

        elif end["EndType"] == "Loop Addendum":
            loop_counter = int(re.search(r"\d+", step_type).group())
            limit["m_szGotoStep"] = step_flow_ctrl[step_index]
            limit["Equation0_szLeft"] = "TC_Counter{}".format(loop_counter)
            limit["Equation0_szCompareSign"] = "<"
            limit["Equation0_szRight"] = end["Value"]

        else:
            raise ValueError("Unable to set end for type {}".format(end["EndType"]))

        return limit

    def convert_report_to_logging_limit(self, report):
        """
        Takes the reporting conditions for the maccor step and converts them to the logging
        limits for the arbin step.

        Args:
            report (OrderedDict): the maccor condition for recording values

        Returns:
            dict: a logging limit that corresponds to the recording conditions for
                maccor report

        """
        ARBIN_SCHEMA = loadfn(
            os.path.join(PROTOCOL_SCHEMA_DIR, "arbin_schedule_schema.yaml")
        )
        limit = ARBIN_SCHEMA["step_blank_limit"]
        limit["m_bStepLimit"] = "0"
        limit["m_bLogDataLimit"] = "1"
        limit["m_szGotoStep"] = "Next Step"

        if report["ReportType"] == "Voltage":
            limit["Equation0_szLeft"] = "DV_Voltage"
            limit["Equation0_szCompareSign"] = ">"
            limit["Equation0_szRight"] = report["Value"]
        elif report["ReportType"] == "Current":
            limit["Equation0_szLeft"] = "DV_Current"
            limit["Equation0_szRight"] = report["Value"]
            limit["Equation0_szCompareSign"] = ">"
        elif report["ReportType"] == "StepTime":
            limit["Equation0_szLeft"] = "DV_Time"
            limit["Equation0_szCompareSign"] = ">"
            if "." in report["Value"]:
                nofrag, frag = report["Value"].split(".")
                frag = frag[:6]  # truncate to microseconds
                frag += (6 - len(frag)) * "0"  # add 0s
                elapsed = datetime.strptime(
                    nofrag.replace("::", "00:00:0"), "%H:%M:%S"
                ).replace(microsecond=int(frag)) - datetime.strptime(
                    "00:00:00", "%H:%M:%S"
                )
            else:
                elapsed = datetime.strptime(
                    report["Value"].replace("::", "00:00:0"), "%H:%M:%S"
                ) - datetime.strptime("00:00:00", "%H:%M:%S")
            limit["Equation0_szRight"] = str(elapsed.total_seconds())

        return limit



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command line/batch interface for structuring many battery cycler runs.

"""

import re

from monty.serialization import loadfn

from beep.conversion_schemas import (
    FastCharge_CONFIG,
    xTesladiag_CONFIG,
    ARBIN_CONFIG,
    MACCOR_CONFIG,
    INDIGO_CONFIG,
    BIOLOGIC_CONFIG,
    NEWARE_CONFIG
)
from beep.structure.arbin import ArbinDatapath
from beep.structure.maccor import MaccorDatapath
from beep.structure.neware import NewareDatapath
from beep.structure.indigo import IndigoDatapath
from beep.structure.biologic import BiologicDatapath
from beep.structure.battery_archive import BatteryArchiveDatapath
from beep.structure.base import BEEPDatapath


def auto_load(filename):
    """Load any supported raw battery cycler file to the correct Datapath automatically.

    Matches raw file patterns to the correct datapath and returns the datapath object.

    Example:
        auto_load("2017-05-09_test-TC-contact_CH33.csv")

        >>> <ArbinDatapath object>

        auto_load("PreDiag_000287_000128short.092")

        >>> <MaccorDatapath object>

    Args:
        filename (str, Pathlike): string corresponding to battery cycler file filename.

    Returns:
        (beep.structure.base.BEEPDatapath): The datapath child class corresponding to this file.

    """
    if re.match(ARBIN_CONFIG["file_pattern"], filename) or re.match(FastCharge_CONFIG["file_pattern"], filename):
        return ArbinDatapath.from_file(filename)
    elif re.match(MACCOR_CONFIG["file_pattern"], filename) or re.match(xTesladiag_CONFIG["file_pattern"], filename):
        return MaccorDatapath.from_file(filename)
    elif re.match(INDIGO_CONFIG["file_pattern"], filename):
        return IndigoDatapath.from_file(filename)
    elif re.match(BIOLOGIC_CONFIG["file_pattern"], filename):
        return BiologicDatapath.from_file(filename)
    elif re.match(NEWARE_CONFIG["file_pattern"], filename):
        return NewareDatapath.from_file(filename)
    elif re.match(BatteryArchiveDatapath.FILE_PATTERN, filename):
        return BatteryArchiveDatapath.from_file(filename)
    else:
        raise ValueError("{} does not match any known file pattern".format(filename))


def auto_load_processed(path):
    """Load processed BEEP .json files regardless of their class.

    Enables loadfn capability for legacy BEEP files, since calling
    loadfn on legacy files will return dictionaries instead of objects
    or will outright fail.

    Examples:

        auto_load_processed("maccor_file_structured.json")

        >>> <MaccorDatapath object>

        auto_load_processed("neware_file_structured.json")

        >>> <NewareDatapath object>

        auto_load_processed("old_Biologic_ProcessedCyclerRun_legacy.json")

        >>> <BiologicDatapath object>

    Args:
        path (str, Pathlike): Path to the structured json file.

    Returns:
        (BEEPDatapath)

    """

    processed_obj = loadfn(path)

    if not isinstance(processed_obj, BEEPDatapath):
        if isinstance(processed_obj, dict):
            # default to Arbin file for BEEPDatapath purposes
            return ArbinDatapath.from_json_file(path)
        else:
            raise TypeError(f"Unknown type for legacy processed json! `{type(processed_obj)}`")
    else:

        # Processed object must have path manually appended
        # as loadfn cannot trigger .from_json_file directly
        processed_obj.paths["structured"] = path
        return processed_obj



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Parsing and conversion of biologic Modulo Bat files to Maccor Procedure Files"""
import os
import re
from copy import deepcopy
from collections import OrderedDict
import xmltodict
from monty.serialization import loadfn
from beep.protocol.biologic import Settings
from beep.protocol.maccor import Procedure
from beep.protocol import PROCEDURE_TEMPLATE_DIR


MACCOR_TEMPLATE = loadfn(os.path.join(PROCEDURE_TEMPLATE_DIR, "maccor_schemas.yaml"))

_default_maccor_header = OrderedDict(MACCOR_TEMPLATE["default_maccor_header"])
_blank_step = OrderedDict(MACCOR_TEMPLATE["blank_step"])
_blank_end_entry = OrderedDict(MACCOR_TEMPLATE["blank_end_entry"])
_blank_report_entry = OrderedDict(MACCOR_TEMPLATE["blank_report_entry"])
_blank_maccor_body = OrderedDict(MACCOR_TEMPLATE["blank_maccor_body"])


class BiologicMbToMaccorProcedure:
    @classmethod
    def convert(
        cls,
        mps_filepath,
        out_filepath,
        mps_file_encoding="ISO-8859-1",
        maccor_header=_default_maccor_header,
    ):
        """
        Biologic Modulo Bat file ingestion from
        *.mps file with a single Modulo Bat technique

        Args:
            mps_filepath(str): *.mps filepath
            out_filepath(str): target destination for Maccor Procedure
            (optional) mps_file_encoding(str): text encoding of *.mps
            (optional) maccor_header(OrderedDict): This defines the shape of the <header> in
                the output file

        Returns: None on success, otherwise throws an error
        """
        if not os.path.isfile(mps_filepath):
            raise Exception(
                "Could not load {}, filepath does not exist".format(mps_filepath)
            )

        with open(mps_filepath, "rb") as mps_file:
            raw_text = mps_file.read()
            text = raw_text.decode(mps_file_encoding)

        protocol_xml = cls.biologic_mb_text_to_maccor_xml(text, maccor_header)

        with open(out_filepath, "w+") as out_file:
            out_file.write(protocol_xml)

        return None

    """
    Biologic Modulo Bat text ingestion from
    *.mps file with a single Modulo Bat technique

    Args:
        text (str): the text from a *.mps file

    Returns:
        (xml): string containing the XML for a Maccor Procedure
    """

    @classmethod
    def biologic_mb_text_to_maccor_xml(
        cls, mps_text, maccor_header=_default_maccor_header
    ):
        schedule_dict = Settings.mps_text_to_schedule_dict(mps_text)

        seqs = cls._get_seqs(schedule_dict)
        steps = cls._create_steps(seqs)

        body = deepcopy(_blank_maccor_body)
        body["MaccorTestProcedure"]["header"] = maccor_header
        body["MaccorTestProcedure"]["ProcSteps"] = OrderedDict({"TestStep": steps})

        newLine = "\r\n"
        xml_with_dirty_empty_elemnts = xmltodict.unparse(
            body, encoding="utf-8", pretty=True, indent="  ", newl=newLine
        )
        xml = Procedure.fixup_empty_elements(xml_with_dirty_empty_elemnts)
        # xmltodict's unparse does not accept custom processing instructions,
        # we must add Maccor's ourselves
        xml_with_maccor_processing_instruction = re.sub(
            r"^[^\r^\n]*",
            '<?xml version="1.0" encoding="UTF-8"?>\r\n<?maccor-application progid="Maccor Procedure File"?>',
            xml,
        )

        # newline at EOF
        return xml_with_maccor_processing_instruction + newLine

    @classmethod
    def _get_seqs(cls, shcedule_dict):
        techniques = shcedule_dict["Technique"]
        assert len(techniques.items()) == 1

        modulo_bat = techniques["1"]
        assert modulo_bat["Type"] == "Modulo Bat"

        # Steps are equivelant to seqs
        numbered_seqs = list(modulo_bat["Step"].items())

        # biologic parser may create empty seq(s) at the end
        # we intentionally _don't_ filter because if there is
        # a blank seq at in the middle
        seqs = list(map(lambda pair: pair[1], numbered_seqs))

        if len(seqs) > 0 and seqs[-1]["Ns"] == "":
            seqs = seqs[:-1]

        return seqs

    @classmethod
    def _stringify_step_num(cls, step_num):
        return "{0:0=3d}".format(step_num)

    # Maccor time can be implemented in
    # [hours]:[minutes]:[seconds]
    #
    # section 4.5.3.4.1 of the maccor manual on Sept, 20, 2020 specifies
    # any combination of numbers may be used as long as the colons
    # for hours minutes and seconds are used e.g.
    # ::.1 = 1 ms
    # ::3600 = 1h or 3600s
    # 05:42:21 = 5h, 42m, 21s
    #
    # no examples of decimals for minutes or hours are given, so that
    # may be a lie.
    #
    # no mention is made of a requirement for leading zeros.
    @staticmethod
    def _convert_time(str_val, unit, field, seq_num):
        if unit == "h":
            return "{}:00:00".format(str_val)
        elif unit == "mn":
            return "00:{}:00".format(str_val)
        elif unit == "s":
            return "00:00:{}".format(str_val)
        elif unit == "ms":
            milliseconds = float(str_val)
            if 0.1 <= milliseconds % 10 <= 9.9:
                raise Exception(
                    "Tried to convert time with specifity at the millisecond level,"
                    + "\nMaccor can only specify up to centiseconds."
                )
            seconds = milliseconds / 1000
            return "00:00:{:.2f}".format(seconds)
        else:
            raise Exception(
                "Unexpected time unit {} for {}, at seq {}".format(unit, field, seq_num)
            )

    # maccor measures only in amps
    @staticmethod
    def _convert_current(str_val, unit, field, seq_num):
        if unit == "A":
            return "{}".format(str_val)
        elif unit == "mA":
            return "{}E-3".format(str_val)
        elif unit in ("\N{Greek Small Letter Mu}A", "\N{Micro Sign}A"):
            return "{}E-6".format(str_val)
        elif unit == "nA":
            return "{}E-9".format(str_val)
        elif unit == "pA":
            return "{}E-12".format(str_val)
        else:
            raise Exception(
                "Unsupported current unit: {} for {}, at seq {}".format(
                    unit, field, seq_num
                )
            )

    @staticmethod
    def _convert_voltage(str_val, unit, field, seq_num):
        if unit == "V":
            return "{}".format(str_val)
        elif unit == "mV":
            return "{}E-3".format(str_val)
        else:
            raise Exception(
                "Unsupported voltage unit: {} for {}, at seq {}".format(
                    unit, field, seq_num
                )
            )

    @staticmethod
    def _convert_resistance(str_val, unit, field, seq_num):
        if unit == "MOhm":
            return "{}E6".format(str_val)
        elif unit == "kOhm":
            return "{}E3".format(str_val)
        elif unit == "Ohm":
            return "{}".format(str_val)
        elif unit == "mOhm":
            return "{}E-3".format(str_val)
        elif unit in ("\N{Greek Small Letter Mu}Ohm", "\N{Micro Sign}Ohm"):
            return "{}E-6".format(str_val)
        else:
            raise Exception(
                "Unsupported resistance unit: {} for {}, at seq {}".format(
                    unit, field, seq_num
                )
            )

    @staticmethod
    def _convert_power(str_val, unit, field, seq_num):
        if unit == "W":
            return "{}".format(str_val)
        elif unit == "mW":
            return "{}E-3".format(str_val)
        elif unit in ("\N{Greek Small Letter Mu}W", "\N{Micro Sign}W"):
            return "{}E-6".format(str_val)
        else:
            raise Exception(
                "Unsupported power unit: {} for {}, at seq {}".format(
                    unit, field, seq_num
                )
            )

    @classmethod
    def _create_loop_step(cls, seq, step_num_by_seq_num):
        step = deepcopy(_blank_step)
        # spacing may be important
        step["StepType"] = " Loop 1 "
        # unclear if spacing is necessary
        step["StepMode"] = "        "

        loop_end_entry = deepcopy(_blank_end_entry)

        loop_count = int(seq["ctrl_repeat"])
        loop_to_seq = int(seq["ctrl_seq"])
        # Maccor requires us to loop to the DO step
        # not the actual procedural step like in Biologic seqs
        loop_to_step = step_num_by_seq_num[loop_to_seq] - 1

        loop_end_entry["EndType"] = "Loop Cnt"
        loop_end_entry["Oper"] = " = "
        loop_end_entry["StepValue"] = "{}".format(loop_count)
        loop_end_entry["Step"] = cls._stringify_step_num(loop_to_step)
        assert len(loop_end_entry["Step"]) == 3

        step["Ends"] = OrderedDict({"EndEntry": loop_end_entry})

        return step

    @classmethod
    def _create_adv_cycle_step(cls):
        # spacing may be important
        step = deepcopy(_blank_step)
        # spacing may be important
        step["StepType"] = "AdvCycle"
        # unclear if spacing is necessary
        step["StepMode"] = "        "
        return step

    @classmethod
    def _create_do_step(cls):
        step = deepcopy(_blank_step)
        # spacing may be important
        step["StepType"] = "  Do 1  "
        # unclear if spacing is necessary
        step["StepMode"] = "        "
        return step

    @classmethod
    def _create_step(
        cls, seq, step_num_by_seq_num, seq_num_is_active_loop_start, end_step_num
    ):
        step = deepcopy(_blank_step)

        seq_num = int(seq["Ns"])
        seq_type = seq["ctrl_type"]
        seq_val = seq["ctrl1_val"]
        seq_unit = seq["ctrl1_val_unit"]
        seq_val_vs = seq["ctrl1_val_vs"]

        step_num = step_num_by_seq_num[seq_num]
        if not isinstance(step_num, int):
            raise Exception(
                "Internal error, could not find step num for seq: {}, this is a bug".format(
                    seq_num
                )
            )

        # all steps except AdvCycle, Do, and Loop seem to have these values
        step["Range"] = "4"
        step["Option1"] = "N"
        step["Option2"] = "N"
        step["Option3"] = "N"

        charge_or_discharge = seq["charge/discharge"]
        if charge_or_discharge != "Charge" and charge_or_discharge != "Discharge":
            raise Exception(
                "unsupported value in charge/discharge field: {}, in seq {}".format(
                    charge_or_discharge, seq_num
                )
            )
        # Dischrge mispelling intentional
        maccor_charge_or_discharge = (
            "Charge " if charge_or_discharge == " Charge " else "Dischrge "
        )

        if seq_type == "CC":
            step["StepMode"] = "Current "
            step["StepType"] = maccor_charge_or_discharge
            step["StepValue"] = cls._convert_current(
                seq_val, seq_unit, "ctrl1_val_vs", seq_num
            )
        elif seq_type == "CV" and seq_val_vs == "Ref":
            step["StepMode"] = "Voltage "
            step["StepType"] = maccor_charge_or_discharge
            step["StepValue"] = cls._convert_voltage(
                seq_val, seq_unit, "ctrl1_val_vs", seq_num
            )
        elif seq_type == "CV":
            raise Exception(
                "Unsupported constant voltage type, ctrl1_val_vs: {} at seq {}".format(
                    seq_val_vs, seq_num
                )
            )
        elif seq_type == "CR":
            step["StepMode"] = "Resistance "
            step["StepType"] = maccor_charge_or_discharge
            step["StepValue"] = cls._convert_resistance(
                seq_val, seq_unit, "ctrl1_val_vs", seq_num
            )
        elif seq_type == "CP":
            step["StepMode"] = "Power "
            step["StepType"] = maccor_charge_or_discharge
            step["StepValue"] = cls._convert_power(
                seq_val, seq_unit, "ctrl1_val_vs", seq_num
            )
        elif seq_type == "Rest":
            step["StepType"] = "  Rest  "
            step["StepMode"] = "        "
        else:
            raise Exception(
                "Unspported ctrl_type: {} at seq_num {}".format(seq_type, seq_num)
            )

        num_limits = int(seq["lim_nb"])
        if num_limits not in [0, 1, 2, 3]:
            raise Exception(
                "Unsupported number of limits, lim_nb: {} at seq {}".format(
                    num_limits, seq_num
                )
                + "\nvalue must in [0, 1, 2, 3]"
            )

        ends = []
        for lim_num in range(1, num_limits + 1):
            end_entry = deepcopy(_blank_end_entry)

            lim_type = seq["lim{}_type".format(lim_num)]
            lim_comp = seq["lim{}_comp".format(lim_num)]
            lim_unit = seq["lim{}_value_unit".format(lim_num)]
            lim_action = seq["lim{}_action".format(lim_num)]
            lim_val = float(seq["lim{}_value".format(lim_num)])
            goto_seq = int(seq["lim{}_seq".format(lim_num)])

            if lim_comp == "<":
                # trailing space intentional
                end_entry["Oper"] = "<= "
            elif lim_comp == ">" and lim_type == "Time":
                # trailing/leading space intentional
                end_entry["Oper"] = " = "
            elif lim_comp == ">":
                # trailing space intentional
                end_entry["Oper"] = ">= "
            else:
                raise Exception(
                    "Unsupported comparator, lim{}_comp: {} at seq {}".format(
                        lim_num, lim_comp, seq_num
                    )
                )

            if lim_type in ("Ece", "Ecell"):
                # trailing space intentional
                end_entry["EndType"] = "Voltage "
                end_entry["Value"] = cls._convert_voltage(
                    lim_val, lim_unit, "lim{}_value_unit".format(lim_num), seq_num
                )
            elif lim_type == "Time":
                # no space intentional
                end_entry["EndType"] = "StepTime"
                end_entry["Value"] = cls._convert_time(
                    lim_val, lim_unit, "lim{}_value_unit".format(lim_num), seq_num
                )
            elif lim_type == "I":
                # trailing space intentional
                end_entry["EndType"] = "Current "
                end_entry["Value"] = cls._convert_current(
                    lim_val, lim_unit, "lim{}_value_unit".format(lim_num), seq_num
                )
            else:
                raise Exception(
                    "Unsupported limit type, lim{}_type: {} at seq {}".format(
                        lim_num, lim_type, seq_num
                    )
                )

            if lim_action == "End":
                end_entry["Step"] = cls._stringify_step_num(end_step_num)
            elif lim_action == "Next sequence":
                next_step = step_num + 1
                end_entry["Step"] = cls._stringify_step_num(next_step)
            elif lim_action == "Goto sequence" and goto_seq not in step_num_by_seq_num:
                raise Exception(
                    "Could not convert goto at seq {}, either lim{}_seq: {}".format(
                        seq_num, lim_num, goto_seq
                    )
                    + "\n{} is not a valid goto seq, or this is a bug".format(goto_seq)
                )
            elif (
                lim_action == "Goto sequence"
                and goto_seq > seq_num
                and goto_seq in seq_num_is_active_loop_start
            ):
                # to go straight to a loop, must go to
                goto_step = step_num_by_seq_num[goto_seq] - 1
                end_entry["Step"] = cls._stringify_step_num(goto_step)
            elif lim_action == "Goto sequence":
                goto_step = step_num_by_seq_num[goto_seq]
                end_entry["Step"] = cls._stringify_step_num(goto_step)
            else:
                raise Exception(
                    "Unsupported goto construct, lim{}_action: {} at seq {}".format(
                        lim_num, lim_action, seq_num
                    )
                )

            ends.append(end_entry)

        assert len(ends) == num_limits
        step["Ends"] = (
            step["Ends"] if len(ends) == 0 else OrderedDict({"EndEntry": ends})
        )

        num_records = int(seq["rec_nb"])
        if num_records not in [0, 1, 2, 3]:
            raise Exception(
                "Unsupported number of records, rec_nb: {} value must in [0, 1, 2, 3]".format(
                    num_records
                )
            )

        reports = []
        for rec_num in range(1, num_records + 1):
            rec_type = seq["rec{}_type".format(rec_num)]
            rec_unit = seq["rec{}_value_unit".format(rec_num)]
            rec_val = float(seq["rec{}_value".format(rec_num)])

            report_entry = deepcopy(_blank_report_entry)

            if rec_type == "Time":
                # no space intentional
                report_entry["ReportType"] = "StepTime"
                report_entry["Value"] = cls._convert_time(
                    rec_val, rec_unit, "rec{}_value_unit".format(rec_num), seq_num
                )
            elif rec_type in ("Ece", "Ecell"):
                # trailing space intentional
                report_entry["ReportType"] = "Voltage "
                report_entry["Value"] = cls._convert_voltage(
                    rec_val, rec_unit, "rec{}_value_unit".format(rec_num), seq_num
                )
            elif rec_type == "I":
                # trailing space intentional
                report_entry["ReportType"] = "Current "
                report_entry["Value"] = cls._convert_current(
                    rec_val, rec_unit, "rec{}_value_unit".format(rec_num), seq_num
                )
            elif rec_type == "Power":
                # trailing space intentional
                report_entry["ReportType"] = "Power "
                report_entry["Value"] = cls._convert_power(
                    rec_val, rec_unit, "rec{}_value_unit".format(rec_num), seq_num
                )
            else:
                raise Exception(
                    "Unsuported record type, rec{}_type: {} at seq {}".format(
                        rec_num, rec_type, seq_num
                    )
                )

            reports.append(report_entry)

        assert len(reports) == num_records
        step["Reports"] = (
            step["Reports"]
            if len(reports) == 0
            else OrderedDict({"ReportEntry": reports})
        )

        return step

    @classmethod
    def _create_steps(cls, seqs):
        # DANGER
        #
        # This code has a lot of complexity and subtlties I couldn't figure out
        # how to remove
        #
        # all of the various passes over the list of sequences are _very_ tightly coupled

        # start build building up a mapping of where loops start and stop
        # since loops are defined at their end instead of beginning, we will
        # iterate from the right.
        #
        # during this process we want to assert loop overlap invariants

        containing_loop_range_by_seq_num = {}
        seq_num_is_active_loop_start = set()
        curr_loop_start = float("inf")
        curr_loop_end = float("inf")

        seqs_in_reverse_order = seqs.copy()
        list.reverse(seqs_in_reverse_order)
        for seq in seqs_in_reverse_order:
            is_loop_seq = seq["ctrl_type"] == "Loop"
            seq_num = int(seq["Ns"])
            loop_to = int(seq["ctrl_seq"])
            # Biologic will pass over this seq without looping
            # If cycle advancement is set to Loop, this seq will not cause the cycle number to advance
            loop_is_inactive = seq["ctrl_repeat"] == "0"

            # handle loops
            if is_loop_seq and not loop_is_inactive and loop_to >= curr_loop_start:
                # Biologic does not have nested loops, maccor does not allow
                # you to loop back into already completed loops
                # to handle this we're disallowing any overlapping loops
                # empty loops that do nothing are excepted
                raise Exception(
                    "Illegal loop at Seq {}. Loops are not allowed to".format(seq_num)
                    + "overlap and Seq {} overlaps with the loop from".format(seq_num)
                    + " seqs {}-{}".format(curr_loop_start, curr_loop_end)
                )
            elif is_loop_seq and not loop_is_inactive:
                curr_loop_end = seq_num
                curr_loop_start = loop_to
                seq_num_is_active_loop_start.update([loop_to])

            if seq_num >= curr_loop_start:
                containing_loop_range_by_seq_num.update(
                    {seq_num: (curr_loop_start, curr_loop_end)}
                )

        for loop_start in seq_num_is_active_loop_start:
            assert loop_start in containing_loop_range_by_seq_num
            assert loop_start == containing_loop_range_by_seq_num[loop_start][0]

        # once we know where loops start and end, we know where to include the
        # "Do" steps, as well as where to create "Adv cycle" steps constructed from
        # a blank loop followed immediately by a loop that goes back to the
        # blank loop exactly once, causing the maccor system to advance the cycle
        #
        # now we'll build up a mapping of seq numbers to step numbers to build out our
        # steps later
        curr_step = 1
        curr_loop_start = -1
        curr_loop_end = -1
        step_num_is_active_loop_start = set()
        seq_num_is_adv_cycle_step = set()
        seq_num_is_blank_loop = set()
        step_num_by_seq_num = {}
        for i, seq in enumerate(seqs):
            seq_num = int(seq["Ns"])
            loop_count = int(seq["ctrl_repeat"])
            loop_to = int(seq["ctrl_seq"])
            is_loop = seq["ctrl_type"] == "Loop"

            if is_loop and loop_count == 0:
                seq_num_is_blank_loop.update([seq_num])
                continue
            elif (
                is_loop
                and loop_to == seq_num - 1
                and (seq_num - 1) in seq_num_is_blank_loop
            ):
                seq_num_is_adv_cycle_step.update([seq_num])
                # add adv cycle step
                step_num_by_seq_num.update({seq_num: curr_step})
                curr_step += 1
                continue

            if seq_num in seq_num_is_active_loop_start:
                curr_loop_start, curr_loop_end = containing_loop_range_by_seq_num[
                    seq_num
                ]
                # add do step
                curr_step += 1
                step_num_is_active_loop_start.update([curr_step])
                seq_num_is_active_loop_start.update([seq_num])

            step_num_by_seq_num[seq_num] = curr_step
            if is_loop and loop_to == i - 1 and (i - 1) in seq_num_is_blank_loop:
                # will adv cycle
                curr_step += 1
            elif is_loop and seqs[loop_to]["ctrl_type"] == "Loop":
                raise Exception(
                    "unhandled looping construct at seq {}, cannot loop to empty loop".format(
                        seq_num
                    )
                    + "\nthat is not immediately followed by a cycle that loops back"
                    + "\nto it (used to force a cycle advance)"
                )
            elif is_loop:
                # will adv cycle and loop
                curr_step += 2
            else:
                # add normal step
                curr_step += 1

        end_step_num = curr_step

        # check for illegal GOTOs
        #
        # Maccor explicitly disallows us to enter a loop from outside a loop (other than to start one)
        # and going back before a loop end is similarly illegal
        steps = []
        last_loop_end = -1
        for seq in seqs:
            seq_num = int(seq["Ns"])
            num_limits = int(seq["lim_nb"])

            if seq_num in seq_num_is_blank_loop:
                continue
            if seq["ctrl_type"] == "Loop":
                last_loop_end = seq_num
                continue

            seq_belongs_to_loop = seq_num in containing_loop_range_by_seq_num

            for lim in range(1, num_limits + 1):
                goto = int(seq["lim{}_seq".format(lim)])
                if goto < last_loop_end:
                    raise Exception(
                        "Illegal GOTO at lim{}_seq, at seq {}\n".format(lim, seq_num)
                        + "\ncannot GOTO any seq that occurs before the end an earlier non-empty loop"
                        + "\nthere was a loop end at seq {}".format(last_loop_end)
                    )

                goto_belongs_to_loop = goto in containing_loop_range_by_seq_num
                goto_is_start_of_loop = (
                    goto == containing_loop_range_by_seq_num[goto][0]
                    if goto_belongs_to_loop
                    else False
                )

                if seq_belongs_to_loop and goto_belongs_to_loop:
                    goto_loop_start, goto_loop_end = containing_loop_range_by_seq_num[
                        goto
                    ]
                    seq_loop_start, seq_loop_end = containing_loop_range_by_seq_num[
                        seq_num
                    ]
                    if (
                        goto_loop_start != seq_loop_start
                        or goto_loop_end != seq_loop_end
                    ):
                        raise Exception(
                            "Illegal GOTO at lim{}_seq: {} at seq {}".format(
                                lim, goto, seq_num
                            )
                            + "\nas {} is contained in loop range {}-{}".format(
                                goto, goto_loop_start, goto_loop_end
                            )
                            + "\nMaccor only allows you to GOTO the interior of a loop from inside the same loop"
                            + "\notherwise you must goto the start of the loop, seq: {}".format(
                                goto_loop_start
                            )
                        )
                if (
                    not seq_belongs_to_loop
                    and goto_belongs_to_loop
                    and not goto_is_start_of_loop
                ):
                    goto_loop_start, goto_loop_end = containing_loop_range_by_seq_num[
                        goto
                    ]
                    raise Exception(
                        "Illegal goto at lim{}_seq: {} at seq {}".format(
                            lim, goto, seq_num
                        )
                        + "\nMaccor only allows you to GOTO the interior of a loop from inside the same loop"
                        + "\nthe loop. Seq {} is in the loop from {}-{} ".format(
                            goto, goto_loop_start, goto_loop_end
                        )
                        + "\nSeq{} is not contained in a loop".format(seq)
                        + "\nYou can only goto the start of the loop, seq {}.".format(
                            goto_loop_start
                        )
                    )

        # tightly coupled with the loop that constructs the seq num/step num mapping
        steps = []
        for i, seq in enumerate(seqs):
            seq_num = int(seq["Ns"])
            seq_is_loop = seq["ctrl_type"] == "Loop"
            loop_to = int(seq["ctrl_seq"])

            if seq_num in seq_num_is_blank_loop:
                continue

            if seq_num in seq_num_is_active_loop_start:
                do_step = cls._create_do_step()
                steps.append(do_step)

            if (
                seq_is_loop
                and loop_to == seq_num - 1
                and seq_num - 1 in seq_num_is_blank_loop
            ):
                adv_cycle_step = cls._create_adv_cycle_step()
                steps.append(adv_cycle_step)
            elif seq_is_loop:
                adv_cycle_step = cls._create_adv_cycle_step()
                steps.append(adv_cycle_step)

                loop_step = cls._create_loop_step(seq, step_num_by_seq_num)
                steps.append(loop_step)
            else:
                step = cls._create_step(
                    seq,
                    step_num_by_seq_num,
                    step_num_is_active_loop_start,
                    end_step_num,
                )
                steps.append(step)

        end_step = deepcopy(_blank_step)
        end_step["StepType"] = "  End   "
        end_step["StepMode"] = "        "
        steps.append(end_step)

        return steps



"""Classes and functions for handling Arbin battery cycler data.

"""
import os
from datetime import datetime

import pytz
from monty.serialization import loadfn
import pandas as pd

from beep.conversion_schemas import ARBIN_CONFIG
from beep import logger, VALIDATION_SCHEMA_DIR
from beep.structure.base import BEEPDatapath
from beep.structure.validate import PROJECT_SCHEMA


class ArbinDatapath(BEEPDatapath):
    """A datapath for Arbin cycler data.

    Arbin cycler data contains two files:

    - Raw data: A raw CSV
    - Metadata: Typically the filename of the raw data + "_Metadata" at the end.

    The metadata file is optional but strongly recommended.

    Attributes:
        All from BEEPDatapath
    """

    conversion_config = ARBIN_CONFIG

    @classmethod
    def from_file(cls, path, metadata_path=None):
        """Load an Arbin file to a datapath.

        Args:
            path (str, Pathlike): Path to the raw data csv.
            metadata_path (str, None): Path to metadata file, if it
                cannot be inferred from the path of the raw file.

        Returns:
            (ArbinDatapath)
        """
        data = pd.read_csv(path, index_col=0)
        data.rename(str.lower, axis="columns", inplace=True)

        for column, dtype in cls.conversion_config["data_types"].items():
            if column in data:
                if not data[column].isnull().values.any():
                    data[column] = data[column].astype(dtype)

        data.rename(cls.conversion_config["data_columns"], axis="columns", inplace=True)

        metadata_path = metadata_path if metadata_path else path.replace(".csv",
                                                                         "_Metadata.csv")

        if os.path.exists(metadata_path):
            metadata = pd.read_csv(metadata_path)
            metadata.rename(str.lower, axis="columns", inplace=True)
            metadata.rename(cls.conversion_config["metadata_fields"], axis="columns",
                            inplace=True)
            # Note the to_dict, which scrubs numpy typing
            metadata = {col: item[0] for col, item in
                        metadata.to_dict("list").items()}
        else:
            logger.warning(f"No associated metadata file for Arbin: "
                           f"'{metadata_path}'. No metadata loaded.")
            metadata = {}

        # standardizing time format
        data["date_time_iso"] = data["date_time"].apply(
            lambda x: datetime.utcfromtimestamp(x).replace(
                tzinfo=pytz.UTC).isoformat()
        )

        paths = {
            "raw": path,
            "metadata": metadata_path if metadata else None
        }

        # Set schema from filename, if possible; otherwise, use default arbin schema
        project_schema = loadfn(PROJECT_SCHEMA)
        name = os.path.basename(path)
        special_schema_filename = project_schema.get(name.split("_")[0], {}).get("arbin")

        if special_schema_filename:
            schema = os.path.join(VALIDATION_SCHEMA_DIR, special_schema_filename)
        else:
            schema = os.path.join(VALIDATION_SCHEMA_DIR, "schema-arbin-lfp.yaml")

        return cls(data, metadata, paths=paths, schema=schema)



"""
Module for finding parameters for projects
"""
import pandas as pd
from glob import glob
import os
from beep import logger


def get_project_sequence(path):
    """
    Returns project sequence for a given path

    Args:
        path (str): full project file path

    Returns:
        ([str]): list of project parts

    """
    root, file = os.path.split(path)
    file = file.split(".")[0]
    file_parts = file.split("_")
    return file_parts


def get_protocol_parameters(filepath, parameters_path):
    """
    Helper function to get the project parameters for a file given the filename

    Args:
        filepath (str): full path to the file
        parameters_path (str): location to look for parameter files

    Returns:
        pandas.DataFrame: single row DataFrame corresponding to the parameters for this file
        pandas.DataFrame: DataFrame with all of the parameter for the project

    """
    project_name_list = get_project_sequence(filepath)
    project_name = project_name_list[0]
    path = os.path.abspath(parameters_path)
    project_parameter_files = glob(os.path.join(path, project_name + "*"))
    assert len(project_parameter_files) <= 1, (
        "Found too many parameter files for: " + project_name
    )

    if len(project_parameter_files) == 1:
        df = pd.read_csv(project_parameter_files[0])
        parameter_row = df[df.seq_num == int(project_name_list[1])]
        if parameter_row.empty:
            logger.error("Unable to get project parameters for: %s", filepath)
            parameter_row = None
            df = None
    else:
        parameter_row = None
        df = None
    return parameter_row, df


def get_diagnostic_parameters(
    diagnostic_available, diagnostic_parameter_path, project_name
):
    """
    Interpolates data according to location and type of diagnostic
    cycles in the data

    Args:
        diagnostic_available (dict): dictionary with diagnostic_types as list,
            'length' of the diagnostic in cycles and location of the diagnostic
        diagnostic_parameter_path (str): full path to the location of the
            diagnostic parameter files
        project_name (str): name of the project to search with

    Returns:
        (list): containing upper and lower voltage limits for the
            diagnostic cycle

    """
    project_diag_files = glob(
        os.path.join(diagnostic_parameter_path, project_name + "*")
    )
    assert len(project_diag_files) <= 1, (
        "Found too many diagnostic parameter files for: " + project_name
    )

    # Find the voltage range for the diagnostic cycles
    if len(project_diag_files) == 1:
        df = pd.read_csv(project_diag_files[0])
        diag_row = df[
            df.diagnostic_parameter_set == diagnostic_available["parameter_set"]
        ]
        v_range = [
            diag_row["diagnostic_discharge_cutoff_voltage"].iloc[0],
            diag_row["diagnostic_charge_cutoff_voltage"].iloc[0],
        ]
    else:
        v_range = [2.7, 4.2]

    return v_range



#  Copyright (c) 2019 Toyota Research Institute

import os

from beep import CONVERSION_SCHEMA_DIR
from monty.serialization import loadfn


ARBIN_CONFIG = loadfn(os.path.join(CONVERSION_SCHEMA_DIR, "arbin_conversion.yaml"))
MACCOR_CONFIG = loadfn(os.path.join(CONVERSION_SCHEMA_DIR, "maccor_conversion.yaml"))
FastCharge_CONFIG = loadfn(
    os.path.join(CONVERSION_SCHEMA_DIR, "FastCharge_conversion.yaml")
)
xTesladiag_CONFIG = loadfn(
    os.path.join(CONVERSION_SCHEMA_DIR, "xTESLADIAG_conversion.yaml")
)
INDIGO_CONFIG = loadfn(os.path.join(CONVERSION_SCHEMA_DIR, "indigo_conversion.yaml"))
NEWARE_CONFIG = loadfn(os.path.join(CONVERSION_SCHEMA_DIR, "neware_conversion.yaml"))
BIOLOGIC_CONFIG = loadfn(
    os.path.join(CONVERSION_SCHEMA_DIR, "biologic_conversion.yaml")
)
MACCOR_WAVEFORM_CONFIG = loadfn(
    os.path.join(CONVERSION_SCHEMA_DIR, "maccor_waveform_conversion.yaml")
)

STRUCTURE_DTYPES = loadfn(os.path.join(CONVERSION_SCHEMA_DIR, "structured_dtypes.yaml"))

ALL_CONFIGS = [ARBIN_CONFIG, MACCOR_CONFIG]



#!/usr/bin/env python3
#  Copyright (c) 2019 Toyota Research Institute

"""Script for re-triggering all of the files at a particular S3 location for processing.
The script produces events that mimic the events from the S3Syncer lambda function,
as a result the same actions should be taken on these files as if they were newly uploaded

Usage:
    retrigger.py [options]
    retrigger.py (-h | --help)

Options:
    -h --help                Show this screen
    --version                Show version
    --mode <mode>            Mode to run in [default: 'test']
    --s3_prefix <s3_prefix>  Prefix to use [default: 'd3Batt/raw/arbin']
    --s3_output <s3_output>  Output prefix to use [default: 'd3Batt/structure']

"""

import time
import datetime
import pytz
import boto3
import collections
import ast
from docopt import docopt
from beep.utils import KinesisEvents

S3_BUCKET_IN = "beep-input-data-stage"
S3_BUCKET_OUT = "beep-output-data-stage"


class DotDict(collections.OrderedDict):
    """Ordered dictionary that can reference keys with "dict.key" notation"""

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, val):
        if key in self.__dict__:
            self.__dict__[key] = val
        else:
            self[key] = val


def eval_args(args):
    """Evaluate docopt arguments and get correctly typed objects

    Args:
        args (dict): Dictionary of arguments parsed by docopt.
    """

    def _fix_boolean(x):
        if x == "True":
            return True
        elif x == "False":
            return False
        else:
            return x

    def _parse_args(x):
        x = _fix_boolean(x)
        try:
            return ast.literal_eval(x)
        except ValueError:
            return x
        except SyntaxError:
            return x

    return DotDict(
        dict(zip(map(lambda x: x[2:], args.keys()), map(_parse_args, args.values())))
    )


def get_structure_name(object):
    file_name = object["Key"].split("/")[-1]
    structure_name = file_name.split(".")[0] + "_structure.json"
    return structure_name


def scan(config):
    print("scanning")
    s3 = boto3.client("s3")
    all_objects = s3.list_objects_v2(Bucket=S3_BUCKET_IN, Prefix=config.s3_prefix)

    objects = [obj for obj in all_objects["Contents"] if obj["Size"] > 1000]

    objects = [
        obj
        for obj in objects
        if "PredictionDiagnostics" in obj["Key"]
        and "x" not in obj["Key"]
        and "Complete" not in obj["Key"]
        # and obj['LastModified'] < datetime.datetime(2020, 3, 24, 5, 35, 43, tzinfo=tzutc())
        # and "_000175_" in obj['Key']
    ]

    old_objects = []
    old = datetime.datetime.now(pytz.utc) - datetime.timedelta(hours=6)
    for obj in objects:
        name = config.s3_output + "/" + get_structure_name(obj)
        structure_objects = s3.list_objects_v2(Bucket=S3_BUCKET_OUT, Prefix=name)
        # print(structure_objects)
        if (
            "Contents" in structure_objects.keys()
            and len(structure_objects["Contents"]) == 1
        ):
            if structure_objects["Contents"][0]["LastModified"] < old:
                old_objects.append(obj)
        else:
            old_objects.append(obj)

    objects = old_objects
    print(len(objects))

    events = KinesisEvents(service="S3Syncer", mode=config.mode)
    objects.reverse()
    for obj in objects:
        retrigger_data = {
            "filename": obj["Key"],
            "bucket": S3_BUCKET_IN,
            "size": obj["Size"],
            "hash": obj["ETag"].strip('"'),
        }
        events.put_upload_retrigger_event("complete", retrigger_data)
        print(retrigger_data)
        time.sleep(0.1)


if __name__ == "__main__":
    docopt_args = docopt(__doc__, version="BEEP Re-Trigger")
    parsed_args = eval_args(docopt_args)
    scan(parsed_args)
